import json
import os
import numpy as np
import time

try:
    import ray
    import transformers
    from transformers import TrainingArguments
    import datasets
    from .dataset.task_auto import get_default_task
    from .result_analysis.azure_utils import JobID
    from .huggingface.trainer import TrainerForAutoTransformers
except ImportError:
    print("To use the nlp component in flaml, run pip install flaml[nlp]")

task_list = [
    "seq-classification",
    "regression",
    "question-answering"
]


class AutoTransformers:
    '''The AutoTransformers class

    Example:

        .. code-block:: python

            autohf = AutoTransformers()
            autohf_settings = {"resources_per_trial": {"cpu": 1},
                       "num_samples": -1,
                       "time_budget": 100000,
                       "ckpt_per_epoch": 1,
                       "fp16": False,
                      }

            validation_metric, analysis = autohf.fit(**autohf_settings)

    '''

    @staticmethod
    def _convert_dict_to_ray_tune_space(config_json, mode="grid"):
        search_space = {}

        if mode == "grid":
            for each_hp in config_json.keys():
                this_config = config_json[each_hp]
                assert isinstance(this_config, dict) or isinstance(this_config, list), \
                    "config of " + each_hp + " must be dict or list"
                search_space[each_hp] = ray.tune.grid_search(this_config)
        else:
            for each_hp in config_json.keys():
                this_config = config_json[each_hp]
                assert isinstance(this_config, dict) or isinstance(this_config, list), \
                    "config of " + each_hp + " must be dict or list"
                if isinstance(this_config, dict):
                    lower = this_config["l"]
                    upper = this_config["u"]
                    space = this_config["space"]
                    if space == "log":
                        search_space[each_hp] = ray.tune.loguniform(lower, upper)
                    elif space == "linear":
                        search_space[each_hp] = ray.tune.uniform(lower, upper)
                    elif space == "quniform":
                        search_space[each_hp] = ray.tune.quniform(lower, upper, this_config["interval"])
                else:
                    search_space[each_hp] = ray.tune.choice(this_config)

        return search_space

    def _set_search_space(self,
                          **custom_hpo_args):
        from .hpo.hpo_searchspace import AutoHPOSearchSpace

        search_space_hpo_json \
            = AutoHPOSearchSpace.from_model_and_dataset_name(self.jobid_config.spa,
                                                             self.jobid_config.pre,
                                                             self.jobid_config.presz,
                                                             self.jobid_config.dat,
                                                             self.jobid_config.subdat,
                                                             **custom_hpo_args)
        self._search_space_hpo = AutoTransformers._convert_dict_to_ray_tune_space(
            search_space_hpo_json,
            mode=self.jobid_config.mod)

    @staticmethod
    def _wrapper(func, *args):  # with star
        return func(*args)

    @staticmethod
    def _get_split_name(data_raw, fold_name=None):
        if fold_name:
            return fold_name
        fold_keys = data_raw.keys()
        if fold_keys == {"train", "validation", "test"}:
            return "train", "validation", "test"
        for each_key in fold_keys:
            for each_split_name in {"train", "validation", "test"}:
                assert not (each_key.startswith(each_split_name) and each_key != each_split_name), \
                    "Dataset split must be within {}, must be explicitly specified in dataset_config, e.g.," \
                    "'fold_name': ['train','validation_matched','test_matched']. Please refer to the example in the " \
                    "documentation of AutoTransformers.prepare_data()".format(",".join(fold_keys))
        return "train", "validation", "test"

    def prepare_data(self,
                     data_root_path,
                     jobid_config=None,
                     is_wandb_on=False,
                     server_name=None,
                     max_seq_length=128,
                     fold_name=None,
                     resplit_portion=None,
                     **custom_data_args):
        """Prepare data

            Example:

                .. code-block:: python

                    preparedata_setting = {"server_name": "tmdev", "data_root_path": "data/", "max_seq_length": 128,
                                               "jobid_config": jobid_config, "wandb_utils": wandb_utils,
                                               "resplit_portion": {"source": ["train", "validation"],
                                               "train": [0, 0.8], "validation": [0.8, 0.9], "test": [0.9, 1.0]}}

                    autohf.prepare_data(**preparedata_setting)

            Args:
                server_name:
                    A string variable, which can be tmdev or azureml
                data_root_path:
                    The root path for storing the checkpoints and output results, e.g., "data/"
                jobid_config:
                    A JobID object describing the profile of job
                wandb_utils:
                    A WandbUtils object for wandb operations
                max_seq_length (optional):
                    Max_seq_lckpt_per_epochength for the huggingface, this hyperparameter must be specified
                    at the data processing step
                resplit_portion:
                    The proportion for resplitting the train and dev data when split_mode="resplit".
                    If args.resplit_mode = "rspt", resplit_portion is required
                is_wandb_on:
                    A boolean variable indicating whether wandb is used
        """
        from .dataset.dataprocess_auto import AutoEncodeText
        from transformers import AutoTokenizer
        from datasets import load_dataset
        from .utils import PathUtils
        from .utils import load_dft_args

        self._max_seq_length = max_seq_length
        self._server_name = server_name if server_name is not None else "tmdev"
        """
            loading the jobid config from console args
        """
        console_args = load_dft_args()
        self.jobid_config = JobID(console_args)
        if jobid_config:
            self.jobid_config = jobid_config
        if len(custom_data_args) > 0:
            self.jobid_config.set_jobid_from_console_args(console_args=custom_data_args)
        if is_wandb_on:
            from .result_analysis.wandb_utils import WandbUtils
            self.wandb_utils = WandbUtils(is_wandb_on=is_wandb_on,
                                          wandb_key_path=console_args.key_path,
                                          jobid_config=self.jobid_config)
            self.wandb_utils.set_wandb_per_run()
        else:
            self.wandb_utils = None

        self.path_utils = PathUtils(self.jobid_config, hpo_data_root_path=data_root_path)

        if self.jobid_config.spt == "rspt":
            assert resplit_portion, "If split mode is 'rspt', the resplit_portion must be provided. Please " \
                                    "refer to the example in the documentation of AutoTransformers.prepare_data()"
        if self.jobid_config.subdat:
            data_raw = load_dataset(JobID.dataset_list_to_str(self.jobid_config.dat),
                                    self.jobid_config.subdat)
        else:
            data_raw = AutoTransformers._wrapper(load_dataset, *self.jobid_config.dat)

        self._train_name, self._dev_name, self._test_name = AutoTransformers._get_split_name(
            data_raw,
            fold_name=fold_name)
        auto_tokentoids_config = {"max_seq_length": self._max_seq_length}
        self._tokenizer = AutoTokenizer.from_pretrained(self.jobid_config.pre_full, use_fast=True)

        def autoencodetext_from_model_and_dataset_name():
            return AutoEncodeText.from_model_and_dataset_name(
                data_raw,
                self.jobid_config.pre_full,
                self.jobid_config.dat,
                self.jobid_config.subdat,
                **auto_tokentoids_config)

        data_encoded = autoencodetext_from_model_and_dataset_name()
        self._max_seq_length = 0
        """
            Update the max_seq_length to the minimum of the actual max seq length and the user defined max_seq_length
        """
        for each_fold in data_encoded.keys():
            self._max_seq_length = max(self._max_seq_length,
                                       max([sum(data_encoded[each_fold][x]['attention_mask']) for x in
                                            range(len(data_encoded[each_fold]))]))
        self._max_seq_length = int((self._max_seq_length + 15) / 16) * 16
        data_encoded = autoencodetext_from_model_and_dataset_name()

        if self.jobid_config.spt == "rspt":
            all_folds_from_source = []
            assert "source" in resplit_portion.keys(), "Must specify the source for resplitting the dataset in" \
                                                       "resplit_portion, which is a list of folder names, e.g., " \
                                                       "resplit_portion = {'source': ['train']}"

            source_fold_names = resplit_portion['source']
            for each_fold_name in source_fold_names:
                this_fold_dataset = data_encoded[each_fold_name]
                all_folds_from_source.append(this_fold_dataset)

            merged_folds_from_source = datasets.concatenate_datasets(all_folds_from_source)
            merged_folds_from_source = merged_folds_from_source.shuffle(seed=self.jobid_config.sddt)

            assert "train" in resplit_portion.keys() and "validation" in resplit_portion.keys() \
                   and "test" in resplit_portion.keys(), "train, validation, test must exist in resplit_portion"

            for key in ["train", "validation", "test"]:
                target_fold_start, target_fold_end = \
                    int(resplit_portion[key][0] * len(merged_folds_from_source)), \
                    int(resplit_portion[key][1] * len(merged_folds_from_source))
                subfold_dataset = merged_folds_from_source.select(
                    [x for x in range(target_fold_start, target_fold_end)]).flatten_indices()
                if key == "train":
                    self.train_dataset = subfold_dataset
                elif key == "validation":
                    self.eval_dataset = subfold_dataset
                else:
                    self.test_dataset = subfold_dataset
        else:
            self.train_dataset, self.eval_dataset, self.test_dataset \
                = data_encoded[self._train_name], data_encoded[self._dev_name], data_encoded[self._test_name]

    def _load_model(self,
                    checkpoint_path=None,
                    per_model_config=None):
        from transformers import AutoConfig
        from .huggingface.switch_head_auto import AutoSeqClassificationHead, MODEL_CLASSIFICATION_HEAD_MAPPING

        this_task = get_default_task(self.jobid_config.dat,
                                     self.jobid_config.subdat)
        if this_task == "seq-classification":
            self._num_labels = len(self.train_dataset.features["label"].names)
        elif this_task == "regression":
            self._num_labels = 1

        if not checkpoint_path:
            checkpoint_path = self.jobid_config.pre_full

        def get_this_model():
            from transformers import AutoModelForSequenceClassification
            return AutoModelForSequenceClassification.from_pretrained(checkpoint_path, config=model_config)

        def is_pretrained_model_in_classification_head_list():
            return self.jobid_config.pre in MODEL_CLASSIFICATION_HEAD_MAPPING.keys()

        def _set_model_config():
            if per_model_config and len(per_model_config) > 0:
                model_config = AutoConfig.from_pretrained(
                    checkpoint_path,
                    num_labels=model_config_num_labels,
                    **per_model_config)
            else:
                model_config = AutoConfig.from_pretrained(
                    checkpoint_path,
                    num_labels=model_config_num_labels)
            return model_config

        if this_task == "seq-classification":
            num_labels_old = AutoConfig.from_pretrained(checkpoint_path).num_labels
            if is_pretrained_model_in_classification_head_list():
                model_config_num_labels = num_labels_old
            else:
                model_config_num_labels = self._num_labels
            model_config = _set_model_config()

            if is_pretrained_model_in_classification_head_list():
                if self._num_labels != num_labels_old:
                    this_model = get_this_model()
                    model_config.num_labels = self._num_labels
                    this_model.num_labels = self._num_labels
                    this_model.classifier = AutoSeqClassificationHead \
                        .from_model_type_and_config(self.jobid_config.pre,
                                                    model_config)
                else:
                    this_model = get_this_model()
            else:
                this_model = get_this_model()

            this_model.resize_token_embeddings(len(self._tokenizer))
            return this_model
        elif this_task == "regression":
            model_config_num_labels = 1
            model_config = _set_model_config()
            this_model = get_this_model()
            return this_model

    def _get_metric_func(self):
        data_name = JobID.dataset_list_to_str(self.jobid_config.dat)
        if data_name in ("glue", "super_glue"):
            metric = datasets.load.load_metric(data_name, self.jobid_config.subdat)
        elif data_name in ("squad", "squad_v2"):
            metric = datasets.load.load_metric(data_name)
        else:
            metric = datasets.load.load_metric(self.metric_name)
        return metric

    def _compute_metrics_by_dataset_name(self,
                                         eval_pred):
        predictions, labels = eval_pred
        predictions = np.squeeze(predictions) \
            if self.task_name == "regression" else np.argmax(predictions, axis=1)
        metric_func = self._get_metric_func()
        return metric_func.compute(predictions=predictions, references=labels)

    def _compute_checkpoint_freq(self,
                                 num_train_epochs,
                                 batch_size):
        if "gpu" in self._resources_per_trial:
            ckpt_step_freq = int(min(num_train_epochs, 1) * len(self.train_dataset) / batch_size
                                 / self._resources_per_trial["gpu"] / self.ckpt_per_epoch) + 1
        else:
            ckpt_step_freq = int(min(num_train_epochs, 1) * len(self.train_dataset) / batch_size
                                 / self._resources_per_trial["cpu"] / self.ckpt_per_epoch) + 1

        return ckpt_step_freq

    @staticmethod
    def _separate_config(config):

        training_args_config = {}
        per_model_config = {}

        for key in config.keys():
            if key in TrainingArguments.__dict__.keys():
                training_args_config[key] = config[key]
            else:
                per_model_config[key] = config[key]

        return training_args_config, per_model_config

    def _objective(self, config, reporter, checkpoint_dir=None):
        from transformers.trainer_utils import set_seed
        self._set_transformers_verbosity(self._transformers_verbose)

        def model_init():
            return self._load_model()
        set_seed(config["seed"])

        training_args_config, per_model_config = AutoTransformers._separate_config(config)
        this_model = self._load_model(per_model_config=per_model_config)

        trial_id = reporter.trial_id
        self.path_utils.make_dir_per_trial(trial_id)

        ckpt_freq = self._compute_checkpoint_freq(
            num_train_epochs=config["num_train_epochs"],
            batch_size=config["per_device_train_batch_size"])

        assert self.path_utils.ckpt_dir_per_trial

        if transformers.__version__.startswith("3"):
            training_args = TrainingArguments(
                output_dir=self.path_utils.ckpt_dir_per_trial,
                do_eval=True,
                per_device_eval_batch_size=32,
                eval_steps=ckpt_freq,
                evaluate_during_training=True,
                save_steps=ckpt_freq,
                save_total_limit=0,
                fp16=self._fp16,
                **training_args_config,
            )
        else:
            from transformers import IntervalStrategy
            training_args = TrainingArguments(
                output_dir=self.path_utils.ckpt_dir_per_trial,
                do_eval=True,
                per_device_eval_batch_size=32,
                eval_steps=ckpt_freq,
                evaluation_strategy=IntervalStrategy.STEPS,
                save_steps=ckpt_freq,
                save_total_limit=0,
                fp16=self._fp16,
                **training_args_config,
            )

        trainer = TrainerForAutoTransformers(
            model=this_model,
            args=training_args,
            model_init=model_init,
            train_dataset=self.train_dataset,
            eval_dataset=self.eval_dataset,
            tokenizer=self._tokenizer,
            compute_metrics=self._compute_metrics_by_dataset_name,
        )
        trainer.trial_id = reporter.trial_id

        """
            create a wandb run. If os.environ["WANDB_MODE"] == "offline", run = None
        """

        if self.wandb_utils:
            run = self.wandb_utils.set_wandb_per_trial()
            import wandb
            for each_hp in config:
                wandb.log({each_hp: config[each_hp]})
        else:
            run = None

        trainer.train()
        trainer.evaluate(self.eval_dataset)
        """
            If a wandb run was created, close the run after train and evaluate finish
        """
        if run:
            run.finish()

    def _verify_init_config(self,
                            **custom_hpo_args):
        for key in custom_hpo_args.keys():
            if key == "points_to_evaluate":
                for each_init_config in custom_hpo_args[key]:
                    for each_hp in each_init_config.keys():
                        assert each_hp in self._search_space_hpo.keys(), \
                            "points_to_evaluate hp must be within the search space"

                        assert isinstance(each_init_config[each_hp], int) or \
                               isinstance(each_init_config[each_hp], float) or \
                               isinstance(each_init_config[each_hp], str) or \
                               isinstance(each_init_config[each_hp], bool), " points_to_evaluate must be a scalar"

                        assert isinstance(self._search_space_hpo[each_hp], ray.tune.sample.Categorical) or \
                               isinstance(self._search_space_hpo[each_hp], ray.tune.sample.Float) or \
                               isinstance(self._search_space_hpo[each_hp], ray.tune.sample.Integer), \
                               "Every hp space must either be categorical, integer or float"

                        if isinstance(self._search_space_hpo[each_hp], ray.tune.sample.Categorical):
                            assert each_init_config[each_hp] in self._search_space_hpo[each_hp].categories, \
                                "points_to_evaluate {each_hp} value must be within the search space"
                        else:
                            assert self._search_space_hpo[each_hp].lower <= each_init_config[each_hp] <= \
                                   self._search_space_hpo[each_hp].upper, \
                                   "points_to_evaluate {each_hp} value must be within the search space"

    def _get_search_algo(self,
                         search_algo_name,
                         search_algo_args_mode,
                         **custom_hpo_args):
        from .hpo.searchalgo_auto import AutoSearchAlgorithm

        if search_algo_name in ("bs", "cfo"):
            self._verify_init_config(**custom_hpo_args)
        search_algo = AutoSearchAlgorithm.from_method_name(
            search_algo_name,
            search_algo_args_mode,
            self._search_space_hpo,
            **custom_hpo_args)
        return search_algo

    @staticmethod
    def _recover_checkpoint(tune_checkpoint_dir):
        assert tune_checkpoint_dir
        # Get subdirectory used for Huggingface.
        subdirs = [
            os.path.join(tune_checkpoint_dir, name)
            for name in os.listdir(tune_checkpoint_dir)
            if os.path.isdir(os.path.join(tune_checkpoint_dir, name))
        ]
        # There should only be 1 subdir.
        assert len(subdirs) == 1, subdirs
        return subdirs[0]

    def _save_ckpt_json(self,
                        best_ckpt):
        json.dump({"best_ckpt": best_ckpt},
                  open(os.path.join(self.path_utils.result_dir_per_run,
                                    "save_ckpt_" + self.jobid_config.to_jobid_string() + ".json"), "w"))

    def _save_output_metric(self,
                            output_metrics):
        json.dump(output_metrics, open(
            os.path.join(self.path_utils.result_dir_per_run,
                         "output_metric_" + self.jobid_config.to_jobid_string() + ".json"), "w"))

    def _load_ckpt_json(self,
                        ckpt_dir=None,
                        **kwargs):
        if not ckpt_dir:
            ckpt_dir = os.path.join(self.path_utils.result_dir_per_run,
                                    "save_ckpt_" + self.jobid_config.to_jobid_string() + ".json")
        try:
            ckpt_json = json.load(open(ckpt_dir))
            return ckpt_json["best_ckpt"]
        except FileNotFoundError as err:
            print("Saved checkpoint not found. Please make sure checkpoint is stored under {}".format(ckpt_dir))
            raise err

    def _set_metric(self, custom_metric_name=None, custom_metric_mode_name=None):
        from .dataset.metric_auto import get_default_and_alternative_metric
        from .utils import _variable_override_default_alternative

        default_metric, default_mode, all_metrics, all_modes = \
            get_default_and_alternative_metric(
                dataset_name_list=self.jobid_config.dat,
                subdataset_name=self.jobid_config.subdat,
                custom_metric_name=custom_metric_name,
                custom_metric_mode_name=custom_metric_mode_name)
        _variable_override_default_alternative(self,
                                               "metric_name",
                                               default_metric,
                                               all_metrics,
                                               custom_metric_name)
        _variable_override_default_alternative(self,
                                               "metric_mode_name",
                                               default_mode,
                                               all_modes,
                                               custom_metric_mode_name)
        self._all_metrics = all_metrics
        self._all_modes = all_modes

    def _set_task(self):
        self.task_name = get_default_task(self.jobid_config.dat,
                                          self.jobid_config.subdat)

    def fit_hf(self,
               resources_per_trial,
               num_samples,
               time_budget,
               custom_metric_name=None,
               custom_metric_mode_name=None,
               _fp16=True,
               **custom_hpo_args
               ):
        from transformers.trainer_utils import HPSearchBackend

        '''Fine tuning the huggingface using HF's API Transformers.hyperparameter_search (for comparitive purpose).
               Transformers.hyperparameter_search has the following disadvantages:
            (1) it does not return tune.analysis.Analysis result, what is analysis used for
            (2) it is inconvenient to develop on top of Transformers.hyperparameter_search, whose trainable function,
                 search space, etc. are defined inside of Transformers.hyperparameter_search.
               An example:
            autohf_settings = {"resources_per_trial": {"cpu": 1},
                       "num_samples": 1,
                       "time_budget": 100000,
                       "ckpt_per_epoch": 1,
                       "fp16": False,
                      }
            validation_metric, analysis = autohf.fit(**autohf_settings,)
            Args:
                resources_per_trial:
                    A dict showing the resources used by each trial,
                    e.g., {"gpu": 4, "cpu": 4}
                num_samples:
                    An int variable of the maximum number of trials
                time_budget:
                    An int variable of the maximum time budget
                custom_metric_name:
                    A string of the dataset name or a function,
                    e.g., 'accuracy', 'f1', 'loss',
                custom_metric_mode_name:
                    A string of the mode name,
                    e.g., "max", "min", "last", "all"
                fp16:
                    boolean, default = True | whether to use fp16
                custom_hpo_args:
                    The additional keyword arguments, e.g.,
                    custom_hpo_args = {"points_to_evaluate": [{
                               "num_train_epochs": 1,
                               "per_device_train_batch_size": 128, }]}
            Returns:
               validation_metric:
                    a dict storing the validation score
            '''

        def model_init():
            return self._load_model()

        def ray_hp_space(trial):
            return {
                "learning_rate": ray.tune.loguniform(1e-6, 1e-4),
                "num_train_epochs": ray.tune.choice(list(range(1, 6))),
                "seed": ray.tune.quniform(1, 41, 1),
                "per_device_train_batch_size": ray.tune.choice([4, 8, 16, 32, 64]),
            }

        self._set_metric(custom_metric_name, custom_metric_mode_name)
        self._set_task()

        training_args = TrainingArguments(
            output_dir=self.path_utils.hpo_ckpt_path,
            fp16=_fp16,
        )
        this_model = self._load_model()

        trainer = TrainerForAutoTransformers(
            this_model,
            training_args,
            model_init=model_init,
            train_dataset=self.train_dataset,
            eval_dataset=self.eval_dataset,
            tokenizer=self._tokenizer,
            compute_metrics=self._compute_metrics_by_dataset_name,
        )
        self.path_utils.make_dir_per_run()

        start_time = time.time()
        best_run = trainer.hyperparameter_search(
            n_trials=num_samples,
            time_budget_s=time_budget,
            # hp_space=ray_hp_space,
            backend=HPSearchBackend.RAY,
            resources_per_trial=resources_per_trial)
        duration = time.time() - start_time
        self.last_run_duration = duration
        print("Total running time: {} seconds".format(duration))

        hp_dict = best_run.hyperparameters
        hp_dict["seed"] = int(hp_dict["seed"])

        best_training_args = TrainingArguments(
            output_dir=self.path_utils.hpo_ckpt_path,
            fp16=_fp16,
            **hp_dict,
        )

        best_trainer = TrainerForAutoTransformers(
            this_model,
            best_training_args,
            model_init=model_init,
            train_dataset=self.train_dataset,
            eval_dataset=self.eval_dataset,
            tokenizer=self._tokenizer,
            compute_metrics=self._compute_metrics_by_dataset_name,
        )

        best_model_checkpoint_path = os.path.join(self.path_utils.hpo_ckpt_path, "hpo_hf")
        if not os.path.exists(best_model_checkpoint_path):
            os.mkdir(best_model_checkpoint_path)
        best_trainer.train()
        best_trainer.save_model(best_model_checkpoint_path)
        self._save_ckpt_json(best_model_checkpoint_path)
        validation_metric = best_trainer.evaluate()

        return validation_metric

    def _set_transformers_verbosity(self, transformers_verbose):
        if transformers_verbose == transformers.logging.ERROR:
            transformers.logging.set_verbosity_error()
        elif transformers_verbose == transformers.logging.WARNING:
            transformers.logging.set_verbosity_warning()
        elif transformers_verbose == transformers.logging.INFO:
            transformers.logging.set_verbosity_info()
        elif transformers_verbose == transformers.logging.DEBUG:
            transformers.logging.set_verbosity_debug()
        else:
            raise Exception("transformers_verbose must be set to ERROR, WARNING, INFO or DEBUG")

    def fit(self,
            num_samples,
            time_budget,
            custom_metric_name=None,
            custom_metric_mode_name=None,
            ckpt_per_epoch=1,
            fp16=True,
            ray_verbose=1,
            transformers_verbose=10,
            resources_per_trial=None,
            ray_local_mode=False,
            **custom_hpo_args):
        """Fine tuning the huggingface using the hpo setting

        Example:

            .. code-block:: python

                autohf_settings = {"resources_per_trial": {"cpu": 1},
                           "num_samples": 1,
                           "time_budget": 100000,
                           "ckpt_per_epoch": 1,
                           "fp16": False,
                          }

                validation_metric, analysis = autohf.fit(**autohf_settings)

        Args:
            resources_per_trial:
                A dict showing the resources used by each trial,
                e.g., {"gpu": 4, "cpu": 4}
            num_samples:
                An int variable of the maximum number of trials
            time_budget:
                An int variable of the maximum time budget
            custom_metric_name:
                A string of the dataset name or a function,
                e.g., 'accuracy', 'f1', 'loss'
            custom_metric_mode_name:
                A string of the mode name,
                e.g., "max", "min", "last", "all"
            ckpt_per_epoch:
                An integer value of number of checkpoints per epoch, default = 1
            ray_verbose:
                An integer, default=1 | verbosit of ray,
            transformers_verbose:
                An integer, default=transformers.logging.INFO | verbosity of transformers, must be chosen from one of
                transformers.logging.ERROR, transformers.logging.INFO, transformers.logging.WARNING,
                or transformers.logging.DEBUG
            fp16:
                A boolean, default = True | whether to use fp16
            ray_local_mode:
                A boolean, default = False | whether to use the local mode (debugging mode) for ray tune.run
            custom_hpo_args:
                The additional keyword arguments, e.g., custom_hpo_args = {"points_to_evaluate": [{
                "num_train_epochs": 1, "per_device_train_batch_size": 128, }]}

        Returns:

            validation_metric: A dict storing the validation score

            analysis: A ray.tune.analysis.Analysis object storing the analysis results from tune.run
        """
        from .hpo.scheduler_auto import AutoScheduler
        self._transformers_verbose = transformers_verbose

        """
         Specify the other parse of jobid configs from custom_hpo_args, e.g., if the search algorithm was not specified
         previously, can specify the algorithm here
        """
        if len(custom_hpo_args) > 0:
            self.jobid_config.set_jobid_from_console_args(console_args=custom_hpo_args)

        self._resources_per_trial = resources_per_trial
        self._set_metric(custom_metric_name, custom_metric_mode_name)
        self._set_task()
        self._fp16 = fp16
        ray.init(local_mode=ray_local_mode)
        self._set_search_space(**custom_hpo_args)

        search_algo = self._get_search_algo(self.jobid_config.alg, self.jobid_config.arg, **custom_hpo_args)
        scheduler = AutoScheduler.from_scheduler_name(self.jobid_config.pru)
        self.ckpt_per_epoch = ckpt_per_epoch
        self.path_utils.make_dir_per_run()

        assert self.path_utils.ckpt_dir_per_run
        start_time = time.time()

        tune_config = self._search_space_hpo
        tune_config["seed"] = self.jobid_config.sdhf

        analysis = ray.tune.run(
            self._objective,
            metric=self.metric_name,
            mode=self.metric_mode_name,
            name="ray_result",
            resources_per_trial=resources_per_trial,
            config=tune_config,
            verbose=ray_verbose,
            local_dir=self.path_utils.ckpt_dir_per_run,
            num_samples=num_samples,
            time_budget_s=time_budget,
            keep_checkpoints_num=1,
            scheduler=scheduler,
            search_alg=search_algo,
        )
        duration = time.time() - start_time
        self.last_run_duration = duration
        print("Total running time: {} seconds".format(duration))

        ray.shutdown()

        best_trial = analysis.get_best_trial(scope="all", metric=self.metric_name, mode=self.metric_mode_name)
        validation_metric = {"eval_" + self.metric_name
                             : best_trial.metric_analysis[self.metric_name][self.metric_mode_name]}
        for x in range(len(self._all_metrics)):
            validation_metric["eval_" + self._all_metrics[x]] \
                = best_trial.metric_analysis[self._all_metrics[x]][self._all_modes[x]]

        get_best_ckpt = analysis.get_best_checkpoint(best_trial, metric=self.metric_name, mode=self.metric_mode_name)
        best_ckpt = AutoTransformers._recover_checkpoint(get_best_ckpt)

        self._save_ckpt_json(best_ckpt)

        return validation_metric, analysis

    def predict(self,
                ckpt_json_dir=None,
                **kwargs):
        '''Predict label for test data.

        An example:
            predictions, test_metric = autohf.predict()

        Args:
            ckpt_json_dir:
                the checkpoint for the fine-tuned huggingface if you wish to override
                the saved checkpoint in the training stage under self.path_utils._result_dir_per_run

        Returns:
            A numpy array of shape n * 1 - - each element is a predicted class
            label for an instance.
        '''
        best_checkpoint = self._load_ckpt_json(ckpt_json_dir, **kwargs)
        best_model = self._load_model(checkpoint_path=best_checkpoint)
        training_args = TrainingArguments(per_device_eval_batch_size=1,
                                          output_dir=self.path_utils.result_dir_per_run)
        test_trainer = TrainerForAutoTransformers(best_model, training_args)

        if self.jobid_config.spt == "ori":
            if "label" in self.test_dataset.features.keys():
                self.test_dataset.remove_columns_("label")
                print("Cleaning the existing label column from test data")

        test_dataloader = test_trainer.get_test_dataloader(self.test_dataset)
        predictions, labels, _ = test_trainer.prediction_loop(test_dataloader, description="Prediction")
        predictions = np.squeeze(predictions) \
            if get_default_task(self.jobid_config.dat,
                                self.jobid_config.subdat) == "regression" \
            else np.argmax(predictions, axis=1)

        if self.jobid_config.spt == "rspt":
            assert labels is not None
            metric = self._get_metric_func()
            output_metric = metric.compute(predictions=predictions, references=labels)
            self._save_output_metric(output_metric)
            return predictions, output_metric
        else:
            return predictions, None

    def output_prediction(self,
                          predictions=None,
                          output_prediction_path=None,
                          output_zip_file_name=None):
        """
            When using the original GLUE split, output the prediction on test data,
            and prepare the .zip file for submission

            Example:
                local_archive_path = self.autohf.output_prediction(predictions,
                                      output_prediction_path= self.console_args.data_root_dir + "result/",
                                      output_zip_file_name=azure_save_file_name)

            Args:
                predictions:
                    A list of predictions, which is the output of AutoTransformers.predict()
                output_prediction_path:
                    Output path for the prediction
                output_zip_file_name:
                    An string, which is the name of the output zip file

            Returns:
                The path of the output .zip file
        """
        from .dataset.submission_auto import auto_output_prediction
        return auto_output_prediction(self.jobid_config.dat,
                                      output_prediction_path,
                                      output_zip_file_name,
                                      predictions,
                                      self.train_dataset,
                                      self._dev_name,
                                      self.jobid_config.subdat)



'''Require: pip install torchvision ray flaml[blendsearch]
'''
import unittest
import os
import time

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.FileHandler('test/tune_pytorch_cifar10.log'))


try:
    import torch
    import torch.nn as nn
    import torch.nn.functional as F
    import torch.optim as optim
    from torch.utils.data import random_split
    import torchvision
    import torchvision.transforms as transforms

    # __net_begin__
    class Net(nn.Module):

        def __init__(self, l1=120, l2=84):
            super(Net, self).__init__()
            self.conv1 = nn.Conv2d(3, 6, 5)
            self.pool = nn.MaxPool2d(2, 2)
            self.conv2 = nn.Conv2d(6, 16, 5)
            self.fc1 = nn.Linear(16 * 5 * 5, l1)
            self.fc2 = nn.Linear(l1, l2)
            self.fc3 = nn.Linear(l2, 10)

        def forward(self, x):
            x = self.pool(F.relu(self.conv1(x)))
            x = self.pool(F.relu(self.conv2(x)))
            x = x.view(-1, 16 * 5 * 5)
            x = F.relu(self.fc1(x))
            x = F.relu(self.fc2(x))
            x = self.fc3(x)
            return x
    # __net_end__
except ImportError:
    print("skip test_pytorch because torchvision cannot be imported.")


# __load_data_begin__
def load_data(data_dir="test/data"):
    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
    ])

    trainset = torchvision.datasets.CIFAR10(
        root=data_dir, train=True, download=True, transform=transform)

    testset = torchvision.datasets.CIFAR10(
        root=data_dir, train=False, download=True, transform=transform)

    return trainset, testset
# __load_data_end__


# __train_begin__
def train_cifar(config, checkpoint_dir=None, data_dir=None):
    if "l1" not in config:
        logger.warning(config)
    net = Net(2 ** config["l1"], 2 ** config["l2"])

    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda:0"
        if torch.cuda.device_count() > 1:
            net = nn.DataParallel(net)
    net.to(device)

    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(net.parameters(), lr=config["lr"], momentum=0.9)

    # The `checkpoint_dir` parameter gets passed by Ray Tune when a checkpoint
    # should be restored.
    if checkpoint_dir:
        checkpoint = os.path.join(checkpoint_dir, "checkpoint")
        model_state, optimizer_state = torch.load(checkpoint)
        net.load_state_dict(model_state)
        optimizer.load_state_dict(optimizer_state)

    trainset, testset = load_data(data_dir)

    test_abs = int(len(trainset) * 0.8)
    train_subset, val_subset = random_split(
        trainset, [test_abs, len(trainset) - test_abs])

    trainloader = torch.utils.data.DataLoader(
        train_subset,
        batch_size=int(2**config["batch_size"]),
        shuffle=True,
        num_workers=4)
    valloader = torch.utils.data.DataLoader(
        val_subset,
        batch_size=int(2**config["batch_size"]),
        shuffle=True,
        num_workers=4)

    from ray import tune

    for epoch in range(int(round(config["num_epochs"]))):  # loop over the dataset multiple times
        running_loss = 0.0
        epoch_steps = 0
        for i, data in enumerate(trainloader, 0):
            # get the inputs; data is a list of [inputs, labels]
            inputs, labels = data
            inputs, labels = inputs.to(device), labels.to(device)

            # zero the parameter gradients
            optimizer.zero_grad()

            # forward + backward + optimize
            outputs = net(inputs)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()

            # print statistics
            running_loss += loss.item()
            epoch_steps += 1
            if i % 2000 == 1999:  # print every 2000 mini-batches
                print("[%d, %5d] loss: %.3f" % (epoch + 1, i + 1,
                                                running_loss / epoch_steps))
                running_loss = 0.0

        # Validation loss
        val_loss = 0.0
        val_steps = 0
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(device), labels.to(device)

                outputs = net(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

                loss = criterion(outputs, labels)
                val_loss += loss.cpu().numpy()
                val_steps += 1

        # Here we save a checkpoint. It is automatically registered with
        # Ray Tune and will potentially be passed as the `checkpoint_dir`
        # parameter in future iterations.
        with tune.checkpoint_dir(step=epoch) as checkpoint_dir:
            path = os.path.join(checkpoint_dir, "checkpoint")
            torch.save(
                (net.state_dict(), optimizer.state_dict()), path)

        tune.report(loss=(val_loss / val_steps), accuracy=correct / total)
    print("Finished Training")
# __train_end__


# __test_acc_begin__
def _test_accuracy(net, device="cpu"):
    trainset, testset = load_data()

    testloader = torch.utils.data.DataLoader(
        testset, batch_size=4, shuffle=False, num_workers=2)

    correct = 0
    total = 0
    with torch.no_grad():
        for data in testloader:
            images, labels = data
            images, labels = images.to(device), labels.to(device)
            outputs = net(images)
            _, predicted = torch.max(outputs.data, 1)
            total += labels.size(0)
            correct += (predicted == labels).sum().item()

    return correct / total
# __test_acc_end__


# __main_begin__
def cifar10_main(
    method='BlendSearch', num_samples=10, max_num_epochs=100, gpus_per_trial=2
):
    data_dir = os.path.abspath("test/data")
    load_data(data_dir)  # Download data for all trials before starting the run
    if method == 'BlendSearch':
        from flaml import tune
    else:
        from ray import tune
    if method in ['BlendSearch', 'BOHB', 'Optuna']:
        config = {
            "l1": tune.randint(2, 8),
            "l2": tune.randint(2, 8),
            "lr": tune.loguniform(1e-4, 1e-1),
            "num_epochs": tune.qloguniform(1, max_num_epochs, q=1),
            "batch_size": tune.randint(1, 4)
        }
    else:
        config = {
            "l1": tune.randint(2, 9),
            "l2": tune.randint(2, 9),
            "lr": tune.loguniform(1e-4, 1e-1),
            "num_epochs": tune.qloguniform(1, max_num_epochs + 1, q=1),
            "batch_size": tune.randint(1, 5)
        }
    import ray
    time_budget_s = 3600
    start_time = time.time()
    if method == 'BlendSearch':
        result = tune.run(
            ray.tune.with_parameters(train_cifar, data_dir=data_dir),
            config=config,
            low_cost_partial_config={
                "l1": 2,
                "l2": 2,
                "num_epochs": 1,
                "batch_size": 4,
            },
            metric="loss",
            mode="min",
            max_resource=max_num_epochs,
            min_resource=1,
            report_intermediate_result=True,
            resources_per_trial={"cpu": 2, "gpu": gpus_per_trial},
            local_dir='logs/',
            num_samples=num_samples,
            time_budget_s=time_budget_s,
            use_ray=True)
    else:
        if 'ASHA' == method:
            algo = None
        elif 'BOHB' == method:
            from ray.tune.schedulers import HyperBandForBOHB
            from ray.tune.suggest.bohb import TuneBOHB
            algo = TuneBOHB()
            scheduler = HyperBandForBOHB(max_t=max_num_epochs)
        elif 'Optuna' == method:
            from ray.tune.suggest.optuna import OptunaSearch
            algo = OptunaSearch()
        elif 'CFO' == method:
            from flaml import CFO
            algo = CFO(low_cost_partial_config={
                "l1": 2,
                "l2": 2,
                "num_epochs": 1,
                "batch_size": 4,
            })
        elif 'Nevergrad' == method:
            from ray.tune.suggest.nevergrad import NevergradSearch
            import nevergrad as ng
            algo = NevergradSearch(optimizer=ng.optimizers.OnePlusOne)
        if method != 'BOHB':
            from ray.tune.schedulers import ASHAScheduler
            scheduler = ASHAScheduler(
                max_t=max_num_epochs,
                grace_period=1)
        result = tune.run(
            tune.with_parameters(train_cifar, data_dir=data_dir),
            resources_per_trial={"cpu": 2, "gpu": gpus_per_trial},
            config=config,
            metric="loss",
            mode="min",
            num_samples=num_samples, time_budget_s=time_budget_s,
            scheduler=scheduler, search_alg=algo
        )
    ray.shutdown()

    logger.info(f"method={method}")
    logger.info(f"n_samples={num_samples}")
    logger.info(f"time={time.time()-start_time}")
    best_trial = result.get_best_trial("loss", "min", "all")
    logger.info("Best trial config: {}".format(best_trial.config))
    logger.info("Best trial final validation loss: {}".format(
        best_trial.metric_analysis["loss"]["min"]))
    logger.info("Best trial final validation accuracy: {}".format(
        best_trial.metric_analysis["accuracy"]["max"]))

    best_trained_model = Net(2**best_trial.config["l1"],
                             2**best_trial.config["l2"])
    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda:0"
        if gpus_per_trial > 1:
            best_trained_model = nn.DataParallel(best_trained_model)
    best_trained_model.to(device)

    checkpoint_path = os.path.join(best_trial.checkpoint.value, "checkpoint")

    model_state, optimizer_state = torch.load(checkpoint_path)
    best_trained_model.load_state_dict(model_state)

    test_acc = _test_accuracy(best_trained_model, device)
    logger.info("Best trial test set accuracy: {}".format(test_acc))
# __main_end__


gpus_per_trial = 0  # 0.5 on GPU server
num_samples = 500


def _test_cifar10_bs():
    cifar10_main(num_samples=num_samples, gpus_per_trial=gpus_per_trial)


def _test_cifar10_cfo():
    cifar10_main('CFO',
                 num_samples=num_samples, gpus_per_trial=gpus_per_trial)


def _test_cifar10_optuna():
    cifar10_main('Optuna',
                 num_samples=num_samples, gpus_per_trial=gpus_per_trial)


def _test_cifar10_asha():
    cifar10_main('ASHA',
                 num_samples=num_samples, gpus_per_trial=gpus_per_trial)


def _test_cifar10_bohb():
    cifar10_main('BOHB',
                 num_samples=num_samples, gpus_per_trial=gpus_per_trial)


def _test_cifar10_nevergrad():
    cifar10_main('Nevergrad',
                 num_samples=num_samples, gpus_per_trial=gpus_per_trial)


if __name__ == "__main__":
    unittest.main()



sentence_keys_glue = {
    "cola": ["sentence"],
    "mnli": ["premise", "hypothesis"],
    "mrpc": ["sentence1", "sentence2"],
    "qnli": ["sentence", "question"],
    "qqp": ["question1", "question2"],
    "rte": ["sentence1", "sentence2"],
    "sst2": ["sentence"],
    "stsb": ["sentence1", "sentence2"],
    "wnli": ["sentence1", "sentence2"]
}

sentence_keys_super_glue = {
    "rte": ["hypothesis", "premise"],
    "wic": ["sentence1", "sentence2"],
    "wsc": ["text"]
}


def get_sentence_keys(dataset_name, subdataset_name=None):
    eval_name_mapping = globals()["sentence_keys_" + dataset_name]
    if isinstance(eval_name_mapping, dict):
        assert subdataset_name and subdataset_name in eval_name_mapping, \
            "dataset_name and subdataset_name not correctly specified"
        sentence_keys = eval_name_mapping[subdataset_name]
    else:
        sentence_keys = eval_name_mapping
    return sentence_keys



'''
Copyright 2020 The Ray Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This source file is adapted here because ray does not fully support Windows.

Copyright (c) Microsoft Corporation.
'''
import uuid
import time
from numbers import Number
from collections import deque
import copy


def flatten_dict(dt, delimiter="/", prevent_delimiter=False):
    dt = copy.deepcopy(dt)
    if prevent_delimiter and any(delimiter in key for key in dt):
        # Raise if delimiter is any of the keys
        raise ValueError(
            "Found delimiter `{}` in key when trying to flatten array."
            "Please avoid using the delimiter in your specification.")
    while any(isinstance(v, dict) for v in dt.values()):
        remove = []
        add = {}
        for key, value in dt.items():
            if isinstance(value, dict):
                for subkey, v in value.items():
                    if prevent_delimiter and delimiter in subkey:
                        # Raise  if delimiter is in any of the subkeys
                        raise ValueError(
                            "Found delimiter `{}` in key when trying to "
                            "flatten array. Please avoid using the delimiter "
                            "in your specification.")
                    add[delimiter.join([key, str(subkey)])] = v
                remove.append(key)
        dt.update(add)
        for k in remove:
            del dt[k]
    return dt


def unflatten_dict(dt, delimiter="/"):
    """Unflatten dict. Does not support unflattening lists."""
    dict_type = type(dt)
    out = dict_type()
    for key, val in dt.items():
        path = key.split(delimiter)
        item = out
        for k in path[:-1]:
            item = item.setdefault(k, dict_type())
        item[path[-1]] = val
    return out


class Trial:
    """A trial object holds the state for one model training run.
    Trials are themselves managed by the TrialRunner class, which implements
    the event loop for submitting trial runs to a Ray cluster.
    Trials start in the PENDING state, and transition to RUNNING once started.
    On error it transitions to ERROR, otherwise TERMINATED on success.
    Attributes:
        trainable_name (str): Name of the trainable object to be executed.
        config (dict): Provided configuration dictionary with evaluated params.
        trial_id (str): Unique identifier for the trial.
        local_dir (str): Local_dir as passed to tune.run.
        logdir (str): Directory where the trial logs are saved.
        evaluated_params (dict): Evaluated parameters by search algorithm,
        experiment_tag (str): Identifying trial name to show in the console.
        resources (Resources): Amount of resources that this trial will use.
        status (str): One of PENDING, RUNNING, PAUSED, TERMINATED, ERROR/
        error_file (str): Path to the errors that this trial has raised.
    """

    PENDING = "PENDING"
    RUNNING = "RUNNING"
    PAUSED = "PAUSED"
    TERMINATED = "TERMINATED"
    ERROR = "ERROR"

    @classmethod
    def generate_id(cls):
        return str(uuid.uuid1().hex)[:8]

    def update_last_result(self, result):
        if self.experiment_tag:
            result.update(experiment_tag=self.experiment_tag)

        self.last_result = result
        self.last_update_time = time.time()

        for metric, value in flatten_dict(result).items():
            if isinstance(value, Number):
                if metric not in self.metric_analysis:
                    self.metric_analysis[metric] = {
                        "max": value,
                        "min": value,
                        "avg": value,
                        "last": value
                    }
                    self.metric_n_steps[metric] = {}
                    for n in self.n_steps:
                        key = "last-{:d}-avg".format(n)
                        self.metric_analysis[metric][key] = value
                        # Store n as string for correct restore.
                        self.metric_n_steps[metric][str(n)] = deque(
                            [value], maxlen=n)
                else:
                    step = result["training_iteration"] or 1
                    self.metric_analysis[metric]["max"] = max(
                        value, self.metric_analysis[metric]["max"])
                    self.metric_analysis[metric]["min"] = min(
                        value, self.metric_analysis[metric]["min"])
                    self.metric_analysis[metric]["avg"] = 1 / step * (
                        value + (step - 1) * self.metric_analysis[metric]["avg"])
                    self.metric_analysis[metric]["last"] = value

                    for n in self.n_steps:
                        key = "last-{:d}-avg".format(n)
                        self.metric_n_steps[metric][str(n)].append(value)
                        self.metric_analysis[metric][key] = sum(
                            self.metric_n_steps[metric][str(n)]) / len(
                                self.metric_n_steps[metric][str(n)])

    def set_status(self, status):
        """Sets the status of the trial."""
        self.status = status
        if status == Trial.RUNNING:
            if self.start_time is None:
                self.start_time = time.time()

    def is_finished(self):
        return self.status in [Trial.ERROR, Trial.TERMINATED]



'''Require: pip install flaml[test,ray]
'''
import time
import os
from sklearn.model_selection import train_test_split
import sklearn.metrics
import sklearn.datasets
try:
    from ray.tune.integration.xgboost import TuneReportCheckpointCallback
except ImportError:
    print("skip test_xgboost because ray tune cannot be imported.")
import xgboost as xgb

import logging
logger = logging.getLogger(__name__)
os.makedirs('logs', exist_ok=True)
logger.addHandler(logging.FileHandler('logs/tune_xgboost.log'))
logger.setLevel(logging.INFO)


def train_breast_cancer(config: dict):
    # This is a simple training function to be passed into Tune
    # Load dataset
    data, labels = sklearn.datasets.load_breast_cancer(return_X_y=True)
    # Split into train and test set
    train_x, test_x, train_y, test_y = train_test_split(
        data, labels, test_size=0.25)
    # Build input matrices for XGBoost
    train_set = xgb.DMatrix(train_x, label=train_y)
    test_set = xgb.DMatrix(test_x, label=test_y)
    # HyperOpt returns a tuple
    config = config.copy()
    config["eval_metric"] = ["logloss", "error"]
    config["objective"] = "binary:logistic"
    # Train the classifier, using the Tune callback
    xgb.train(
        config,
        train_set,
        evals=[(test_set, "eval")],
        verbose_eval=False,
        callbacks=[TuneReportCheckpointCallback(filename="model.xgb")])


def _test_xgboost(method='BlendSearch'):
    try:
        import ray
    except ImportError:
        return
    if method == 'BlendSearch':
        from flaml import tune
    else:
        from ray import tune
    search_space = {
        "max_depth": tune.randint(1, 8) if method in [
            "BlendSearch", "BOHB", "Optuna"] else tune.randint(1, 9),
        "min_child_weight": tune.choice([1, 2, 3]),
        "subsample": tune.uniform(0.5, 1.0),
        "eta": tune.loguniform(1e-4, 1e-1)
    }
    max_iter = 10
    for num_samples in [128]:
        time_budget_s = 60
        for n_cpu in [8]:
            start_time = time.time()
            ray.init(num_cpus=n_cpu, num_gpus=0)
            # ray.init(address='auto')
            if method == 'BlendSearch':
                analysis = tune.run(
                    train_breast_cancer,
                    config=search_space,
                    low_cost_partial_config={
                        "max_depth": 1,
                    },
                    cat_hp_cost={
                        "min_child_weight": [6, 3, 2],
                    },
                    metric="eval-logloss",
                    mode="min",
                    max_resource=max_iter,
                    min_resource=1,
                    report_intermediate_result=True,
                    # You can add "gpu": 0.1 to allocate GPUs
                    resources_per_trial={"cpu": 1},
                    local_dir='logs/',
                    num_samples=num_samples * n_cpu,
                    time_budget_s=time_budget_s,
                    use_ray=True)
            else:
                if 'ASHA' == method:
                    algo = None
                elif 'BOHB' == method:
                    from ray.tune.schedulers import HyperBandForBOHB
                    from ray.tune.suggest.bohb import TuneBOHB
                    algo = TuneBOHB(max_concurrent=n_cpu)
                    scheduler = HyperBandForBOHB(max_t=max_iter)
                elif 'Optuna' == method:
                    from ray.tune.suggest.optuna import OptunaSearch
                    algo = OptunaSearch()
                elif 'CFO' == method:
                    from flaml import CFO
                    algo = CFO(low_cost_partial_config={
                        "max_depth": 1,
                    }, cat_hp_cost={
                        "min_child_weight": [6, 3, 2],
                    })
                elif 'CFOCat' == method:
                    from flaml.searcher.cfo_cat import CFOCat
                    algo = CFOCat(low_cost_partial_config={
                        "max_depth": 1,
                    }, cat_hp_cost={
                        "min_child_weight": [6, 3, 2],
                    })
                elif 'Dragonfly' == method:
                    from ray.tune.suggest.dragonfly import DragonflySearch
                    algo = DragonflySearch()
                elif 'SkOpt' == method:
                    from ray.tune.suggest.skopt import SkOptSearch
                    algo = SkOptSearch()
                elif 'Nevergrad' == method:
                    from ray.tune.suggest.nevergrad import NevergradSearch
                    import nevergrad as ng
                    algo = NevergradSearch(optimizer=ng.optimizers.OnePlusOne)
                elif 'ZOOpt' == method:
                    from ray.tune.suggest.zoopt import ZOOptSearch
                    algo = ZOOptSearch(budget=num_samples * n_cpu)
                elif 'Ax' == method:
                    from ray.tune.suggest.ax import AxSearch
                    algo = AxSearch()
                elif 'HyperOpt' == method:
                    from ray.tune.suggest.hyperopt import HyperOptSearch
                    algo = HyperOptSearch()
                    scheduler = None
                if method != 'BOHB':
                    from ray.tune.schedulers import ASHAScheduler
                    scheduler = ASHAScheduler(
                        max_t=max_iter,
                        grace_period=1)
                analysis = tune.run(
                    train_breast_cancer,
                    metric="eval-logloss",
                    mode="min",
                    # You can add "gpu": 0.1 to allocate GPUs
                    resources_per_trial={"cpu": 1},
                    config=search_space, local_dir='logs/',
                    num_samples=num_samples * n_cpu,
                    time_budget_s=time_budget_s,
                    scheduler=scheduler, search_alg=algo)
            ray.shutdown()
            # # Load the best model checkpoint
            # import os
            # best_bst = xgb.Booster()
            # best_bst.load_model(os.path.join(analysis.best_checkpoint,
            #  "model.xgb"))
            best_trial = analysis.get_best_trial("eval-logloss", "min", "all")
            accuracy = 1. - best_trial.metric_analysis["eval-error"]["min"]
            logloss = best_trial.metric_analysis["eval-logloss"]["min"]
            logger.info(f"method={method}")
            logger.info(f"n_samples={num_samples*n_cpu}")
            logger.info(f"time={time.time()-start_time}")
            logger.info(f"Best model eval loss: {logloss:.4f}")
            logger.info(f"Best model total accuracy: {accuracy:.4f}")
            logger.info(f"Best model parameters: {best_trial.config}")


def test_nested():
    from flaml import tune
    search_space = {
        # test nested search space
        "cost_related": {
            "a": tune.randint(1, 8),
        },
        "b": tune.uniform(0.5, 1.0),
    }

    def simple_func(config):
        obj = (config["cost_related"]["a"] - 4)**2 \
            + (config["b"] - config["cost_related"]["a"])**2
        tune.report(obj=obj)
        tune.report(obj=obj, ab=config["cost_related"]["a"] * config["b"])

    analysis = tune.run(
        simple_func,
        config=search_space,
        low_cost_partial_config={
            "cost_related": {"a": 1}
        },
        metric="obj",
        mode="min",
        metric_constraints=[("ab", "<=", 4)],
        local_dir='logs/',
        num_samples=-1,
        time_budget_s=1)

    best_trial = analysis.get_best_trial()
    logger.info(f"Best config: {best_trial.config}")
    logger.info(f"Best result: {best_trial.last_result}")


def test_xgboost_bs():
    _test_xgboost()


def _test_xgboost_cfo():
    _test_xgboost('CFO')


def test_xgboost_cfocat():
    _test_xgboost('CFOCat')


def _test_xgboost_dragonfly():
    _test_xgboost('Dragonfly')


def _test_xgboost_skopt():
    _test_xgboost('SkOpt')


def _test_xgboost_nevergrad():
    _test_xgboost('Nevergrad')


def _test_xgboost_zoopt():
    _test_xgboost('ZOOpt')


def _test_xgboost_ax():
    _test_xgboost('Ax')


def __test_xgboost_hyperopt():
    _test_xgboost('HyperOpt')


def _test_xgboost_optuna():
    _test_xgboost('Optuna')


def _test_xgboost_asha():
    _test_xgboost('ASHA')


def _test_xgboost_bohb():
    _test_xgboost('BOHB')


if __name__ == "__main__":
    test_xgboost_bs()



from collections import OrderedDict
from functools import partial

from transformers import AutoTokenizer
from .sentence_keys_auto import get_sentence_keys


def inserting_sepp(sent, start, end, this_tokenizer):
    return \
        sent[:start].rstrip() + " " + this_tokenizer.sep_token + " " + sent[start:end] \
        + " " + this_tokenizer.sep_token + " " + sent[end:].lstrip()


def tokenize_superglue_copa(this_example,
                            this_tokenizer,
                            dataset_name,
                            subdataset_name=None,
                            **kwargs):
    return None


def tokenize_superglue_wic_gpt2(this_example,
                                this_tokenizer,
                                dataset_name,
                                subdataset_name=None,
                                **kwargs):
    return None


def tokenize_superglue_wic(this_example,
                           this_tokenizer,
                           dataset_name,
                           subdataset_name=None,
                           **kwargs
                           ):
    """
        tokenize the data from the wic task (word-in-context dataset),
        e.g., sentence 1: "There's a lot of trash on the bed of the river"
        sentence 2: "I keep a glass of water next to my bed when I sleep",
        label = False (different word senses)
        In the superglue data, the position of the word in sentence 1 and 2 are provided
        What this function does is to update the span position after tokenization, based on each LM's own tokenizer,
        The key is to insert an [SEP] before and after the original sentence, then feed it into the LM's tokenizer.
        There are two challenges:
           (1) Each LM's tokenizations are different, e.g., in XLNet's tokenizer, the paddings are on the left'
           (2) Some LM's tokenization would add an underline symbol before the word, e.g., "There's a lot"
           -> [_There, _', _s, _a, _lot]
           When underline meets special char such as '"', "'", the tokenized sequence after adding [SEP] needs to be
           aligned with the sequence tokenized without [SEP]. We use a two pointer algorithm for the alignment
    """
    sent1, sent2 = this_example["sentence1"], this_example["sentence2"]
    start1, end1 = this_example["start1"], this_example["end1"]
    start2, end2 = this_example["start2"], this_example["end2"]
    """
        Add [SEP] to the sentence
    """
    altered_sent1 = inserting_sepp(sent1, start1, end1, this_tokenizer)
    altered_sent2 = inserting_sepp(sent2, start2, end2, this_tokenizer)
    input_ids_sepp = this_tokenizer(*(altered_sent1, altered_sent2),
                                    padding="max_length",
                                    max_length=1024,
                                    truncation=True)["input_ids"]
    data_pair = (sent1, sent2)
    assert "max_seq_length" in kwargs, "max_seq_length must be provided for glue"
    this_data = this_tokenizer(*data_pair, padding="max_length", max_length=kwargs["max_seq_length"], truncation=True)
    input_ids = this_data["input_ids"]
    which_sepp = 0

    """
        span_start_end: a 2x2 array:
        * (span_start_end[0][0], span_start_end[0][1]) are the spans of the position of the word in the first sentence
        * (span_start_end[1][0], span_start_end[1][1]) are the spans of the position of the word in the second sentence
    """
    span_start_end = [[-1, -1], [-1, -1]]

    ptr_sepp = ptr_nosepp = 0
    try:
        padding_direction = this_tokenizer.padding_side
        if padding_direction == "left":
            padding_id = input_ids_sepp[0]
            while input_ids_sepp[ptr_sepp] == padding_id:
                ptr_sepp += 1
            while input_ids[ptr_nosepp] == padding_id:
                ptr_nosepp += 1
    except KeyError:
        pass
    sep_id = this_tokenizer.convert_tokens_to_ids([this_tokenizer.sep_token])[0]
    """
        use two pointers to align the tokenized sequence before and after adding [SEP];
        ptr_sepp: the pointer after adding; ptr_nosepp: the pointer without adding
    """
    while ptr_sepp < len(input_ids_sepp) and ptr_nosepp < len(input_ids) and \
            input_ids_sepp[ptr_sepp] != 0 and input_ids[ptr_nosepp] != 0:
        if input_ids_sepp[ptr_sepp] == input_ids[ptr_nosepp]:
            ptr_sepp += 1
            ptr_nosepp += 1
        else:
            if not (input_ids_sepp[ptr_sepp] == sep_id
                    or this_tokenizer.convert_ids_to_tokens([input_ids_sepp[ptr_sepp]])[0] in ('▁', '_')):
                break
            if input_ids_sepp[ptr_sepp] == sep_id:
                span_start_end[int(which_sepp / 2)][which_sepp % 2] = ptr_nosepp
                which_sepp += 1
                ptr_sepp += 1
            else:
                ptr_sepp += 1
    """
        max_word_span is the maximum tokens of the word
        It is set to 16 following deberta:
        https://github.com/microsoft/DeBERTa/blob/master/DeBERTa/apps/tasks/superglue_tasks.py#L1054
    """
    max_word_span = 16
    word_indices = []
    for idx1 in range(2):
        if span_start_end[idx1][1] < kwargs["max_seq_length"]:
            first_span = [x for x in range(span_start_end[idx1][0], span_start_end[idx1][1])
                          if x < kwargs["max_seq_length"]] + [0] * (max_word_span - span_start_end[idx1][1]
                                                                    + span_start_end[idx1][0])
            word_indices.append(first_span)
    this_data["word_spans"] = word_indices
    return this_data


def tokenize_glue(this_example,
                  this_tokenizer,
                  dataset_name,
                  subdataset_name=None,
                  **kwargs):
    sentence_keys = get_sentence_keys(dataset_name, subdataset_name)

    if len(sentence_keys) > 1:
        sentence1_key, sentence2_key = sentence_keys[0], sentence_keys[1]
    else:
        sentence1_key = sentence_keys[0]
        sentence2_key = None

    data_pair = (
        (this_example[sentence1_key],) if sentence2_key is None else (
            this_example[sentence1_key], this_example[sentence2_key])
    )
    assert "max_seq_length" in kwargs, "max_seq_length must be provided for glue"
    return this_tokenizer(*data_pair, padding="max_length", max_length=kwargs["max_seq_length"], truncation=True)


TOKENIZER_MAPPING = OrderedDict(
    [
        (("glue", "rte"), tokenize_glue),
        (("glue", "mrpc"), tokenize_glue),
        (("glue", "cola"), tokenize_glue),
        (("glue", "wnli"), tokenize_glue),
        (("glue", "stsb"), tokenize_glue),
        (("glue", "sst2"), tokenize_glue),
        (("glue", "mnli"), tokenize_glue),
        (("glue", "qqp"), tokenize_glue),
        (("glue", "qnli"), tokenize_glue),
        (("super_glue", "wic"), tokenize_superglue_wic),
    ]
)


class AutoEncodeText:
    """
    This is a generic input text tokenization class that will be instantiated as one of the
    tokenization classes of the library when created with the
    `~flaml.nlp.dataset.AutoEncodeText.from_model_and_dataset_name` class method.

    This class cannot be instantiated directly using ``__init__()`` (throws an error).
    """

    def __init__(self):
        raise EnvironmentError(
            "AutoEncodeText is designed to be instantiated "
            "using the `AutoEncodeText.from_model_and_dataset_name(cls,"
            "data_raw,model_checkpoint_path,dataset_name,subdataset_name = None,**kwargs)` methods."
        )

    @classmethod
    def from_model_and_dataset_name(cls,
                                    data_raw,
                                    model_checkpoint_path,
                                    dataset_name_list: list = None,
                                    subdataset_name=None,
                                    **kwargs):
        """
        Instantiate one of the input text tokenization classes from the raw data, model checkpoint path, dataset name
        and sub dataset name. The raw data is used for creating a mapping function from the raw tokens to the
        tokenized token ids.

        Args:
            data_raw:
                The raw data (a datasets.Dataset object)

            model_checkpoint_path:
                A string variable which specifies the model path, e.g., "google/electra-base-discriminator"

            dataset_name_list:
                A list which is the dataset name, e.g., ["glue"]

            subdataset_name:
                A string variable which is the sub dataset name,e.g., "rte"

            kwargs:
                The values in kwargs of any keys will be used for the mapping function

        Examples:
            >>> from datasets import load_dataset
            >>> data_raw = load_dataset("glue", "rte")
            >>> AutoEncodeText.from_model_and_dataset_name(data_raw, "google/electra-base-discriminator", ["glue"], "rte")

        """
        from ..result_analysis.azure_utils import JobID
        dataset_name = JobID.dataset_list_to_str(dataset_name_list)
        if (dataset_name, subdataset_name) in TOKENIZER_MAPPING.keys():
            this_tokenizer = AutoTokenizer.from_pretrained(model_checkpoint_path, use_fast=True)
            token_func = TOKENIZER_MAPPING[(dataset_name, subdataset_name)]
            return data_raw.map(
                partial(token_func,
                        this_tokenizer=this_tokenizer,
                        dataset_name=dataset_name,
                        subdataset_name=subdataset_name,
                        **kwargs), batched=False)
        raise ValueError(
            "Unrecognized method {},{} for this kind of AutoGridSearchSpace: {}.\n"
            "Method name should be one of {}.".format(
                dataset_name, subdataset_name, cls.__name__, ", ".join(c[0] for c in TOKENIZER_MAPPING.keys())
            )
        )



from collections import OrderedDict
from ray.tune.schedulers import ASHAScheduler, HyperBandScheduler

SCHEDULER_MAPPING = OrderedDict(
    [
        ("None", None),
        ("asha", ASHAScheduler),
        ("hb", HyperBandScheduler),
    ]
)


class AutoScheduler:
    """
    This is a class for getting the scheduler based on the scheduler name
    (a string variable) instantiated as one of the schedulers of the library when
    created with the `~flaml.nlp.hpo.AutoScheduler.from_scheduler_name` method.

    This class cannot be instantiated directly using ``__init__()`` (throws an error).
    """

    def __init__(self):
        raise EnvironmentError(
            "AutoScheduler is designed to be instantiated "
            "using the `AutoScheduler.from_scheduler_name(cls, scheduler_name, **kwargs)` methods."
        )

    @classmethod
    def from_scheduler_name(cls, scheduler_name, **kwargs):
        """
        Instantiate one of the schedulers using the scheduler names

        Args:
            scheduler_name:
                A string variable for the scheduler name

        Example:
            >>> AutoScheduler.from_scheduler_name("asha")
        """
        if scheduler_name in SCHEDULER_MAPPING.keys():
            if SCHEDULER_MAPPING[scheduler_name] is None:
                return None
            return SCHEDULER_MAPPING[scheduler_name](**kwargs)
        raise ValueError(
            "Unrecognized scheduler {} for this kind of AutoScheduler: {}.\n"
            "Scheduler name should be one of {}.".format(
                scheduler_name, cls.__name__, ", ".join(SCHEDULER_MAPPING.keys())
            )
        )



import os
import unittest
from tempfile import TemporaryDirectory

from sklearn.datasets import load_boston

from flaml import AutoML
from flaml.training_log import training_log_reader


class TestTrainingLog(unittest.TestCase):

    def test_training_log(self):

        with TemporaryDirectory() as d:
            filename = os.path.join(d, 'test_training_log.log')

            # Run a simple job.
            automl_experiment = AutoML()
            automl_settings = {
                "time_budget": 2,
                "metric": 'mse',
                "task": 'regression',
                "log_file_name": filename,
                "log_training_metric": True,
                "mem_thres": 1024 * 1024,
                "n_jobs": 1,
                "model_history": True,
                "train_time_limit": 0.01,
                "verbose": 3,
                "ensemble": True,
            }
            X_train, y_train = load_boston(return_X_y=True)
            automl_experiment.fit(X_train=X_train, y_train=y_train,
                                  **automl_settings)

            # Check if the training log file is populated.
            self.assertTrue(os.path.exists(filename))
            with training_log_reader(filename) as reader:
                count = 0
                for record in reader.records():
                    print(record)
                    count += 1
                self.assertGreater(count, 0)



'''!
 * Copyright (c) 2020-2021 Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License. See LICENSE file in the
 * project root for license information.
'''
from typing import Dict, Optional
import numpy as np
try:
    from ray.tune.suggest import Searcher
except ImportError:
    from .suggestion import Searcher
from .flow2 import FLOW2

import logging
logger = logging.getLogger(__name__)


class SearchThread:
    '''Class of global or local search thread
    '''

    _eps = 1.0

    def __init__(self, mode: str = "min",
                 search_alg: Optional[Searcher] = None,
                 cost_attr: Optional[str] = 'time_total_s'):
        ''' When search_alg is omitted, use local search FLOW2
        '''
        self._search_alg = search_alg
        self._is_ls = isinstance(search_alg, FLOW2)
        self._mode = mode
        self._metric_op = 1 if mode == 'min' else -1
        self.cost_best = self.cost_last = self.cost_total = self.cost_best1 = \
            getattr(search_alg, 'cost_incumbent', 0)
        self.cost_best2 = 0
        self.obj_best1 = self.obj_best2 = getattr(
            search_alg, 'best_obj', np.inf)  # inherently minimize
        # eci: estimated cost for improvement
        self.eci = self.cost_best
        self.priority = self.speed = 0
        self._init_config = True
        self.running = 0    # the number of running trials from the thread
        self.cost_attr = cost_attr

    @classmethod
    def set_eps(cls, time_budget_s):
        cls._eps = max(min(time_budget_s / 1000.0, 1.0), 1e-10)

    def suggest(self, trial_id: str) -> Optional[Dict]:
        ''' use the suggest() of the underlying search algorithm
        '''
        if isinstance(self._search_alg, FLOW2):
            config = self._search_alg.suggest(trial_id)
        else:
            try:
                config = self._search_alg.suggest(trial_id)
            except FloatingPointError:
                logger.warning(
                    'The global search method raises FloatingPointError. '
                    'Ignoring for this iteration.')
                config = None
        if config is not None:
            self.running += 1
        return config

    def update_priority(self, eci: Optional[float] = 0):
        # optimistic projection
        self.priority = eci * self.speed - self.obj_best1

    def update_eci(self, metric_target: float,
                   max_speed: Optional[float] = np.inf):
        # calculate eci: estimated cost for improvement over metric_target
        best_obj = metric_target * self._metric_op
        if not self.speed:
            self.speed = max_speed
        self.eci = max(self.cost_total - self.cost_best1,
                       self.cost_best1 - self.cost_best2)
        if self.obj_best1 > best_obj and self.speed > 0:
            self.eci = max(self.eci, 2 * (self.obj_best1 - best_obj) / self.speed)

    def _update_speed(self):
        # calculate speed; use 0 for invalid speed temporarily
        if self.obj_best2 > self.obj_best1:
            # discount the speed if there are unfinished trials
            self.speed = (self.obj_best2 - self.obj_best1) / self.running / (
                max(self.cost_total - self.cost_best2, SearchThread._eps))
        else:
            self.speed = 0

    def on_trial_complete(self, trial_id: str, result: Optional[Dict] = None,
                          error: bool = False):
        ''' update the statistics of the thread
        '''
        if not self._search_alg:
            return
        if not hasattr(self._search_alg, '_ot_trials') or (
                not error and trial_id in self._search_alg._ot_trials):
            # optuna doesn't handle error
            if self._is_ls or not self._init_config:
                try:
                    self._search_alg.on_trial_complete(trial_id, result, error)
                except RuntimeError as e:
                    # rs is used in place of optuna sometimes
                    if not str(e).endswith(
                       "has already finished and can not be updated."):
                        raise e
            else:
                # init config is not proposed by self._search_alg
                # under this thread
                self._init_config = False
        if result:
            self.cost_last = result.get(self.cost_attr, 1)
            self.cost_total += self.cost_last
            if self._search_alg.metric in result:
                obj = result[self._search_alg.metric] * self._metric_op
                if obj < self.obj_best1:
                    self.cost_best2 = self.cost_best1
                    self.cost_best1 = self.cost_total
                    self.obj_best2 = obj if np.isinf(
                        self.obj_best1) else self.obj_best1
                    self.obj_best1 = obj
                    self.cost_best = self.cost_last
            self._update_speed()
        self.running -= 1
        assert self.running >= 0

    def on_trial_result(self, trial_id: str, result: Dict):
        ''' TODO update the statistics of the thread with partial result?
        '''
        if not self._search_alg:
            return
        if not hasattr(self._search_alg, '_ot_trials') or (
           trial_id in self._search_alg._ot_trials):
            try:
                self._search_alg.on_trial_result(trial_id, result)
            except RuntimeError as e:
                # rs is used in place of optuna sometimes
                if not str(e).endswith(
                   "has already finished and can not be updated."):
                    raise e
        if self.cost_attr in result and self.cost_last < result[self.cost_attr]:
            self.cost_last = result[self.cost_attr]
            # self._update_speed()

    @property
    def converged(self) -> bool:
        return self._search_alg.converged

    @property
    def resource(self) -> float:
        return self._search_alg.resource

    def reach(self, thread) -> bool:
        ''' whether the incumbent can reach the incumbent of thread
        '''
        return self._search_alg.reach(thread._search_alg)

    @property
    def can_suggest(self) -> bool:
        ''' whether the thread can suggest new configs
        '''
        return self._search_alg.can_suggest



'''!
 * Copyright (c) 2020-2021 Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License. See LICENSE file in the
 * project root for license information.
'''
from typing import Dict, Optional, List, Tuple, Callable
import numpy as np
import time
import pickle

try:
    from ray.tune.suggest import Searcher
    from ray.tune.suggest.optuna import OptunaSearch as GlobalSearch
    from ray.tune.suggest.variant_generator import generate_variants
except ImportError:
    from .suggestion import Searcher
    from .suggestion import OptunaSearch as GlobalSearch
    from .variant_generator import generate_variants
from .search_thread import SearchThread
from .flow2 import FLOW2

import logging
logger = logging.getLogger(__name__)


class BlendSearch(Searcher):
    '''class for BlendSearch algorithm
    '''

    cost_attr = "time_total_s"  # cost attribute in result
    lagrange = '_lagrange'      # suffix for lagrange-modified metric
    penalty = 1e+10             # penalty term for constraints
    LocalSearch = FLOW2

    def __init__(self,
                 metric: Optional[str] = None,
                 mode: Optional[str] = None,
                 space: Optional[dict] = None,
                 points_to_evaluate: Optional[List[dict]] = None,
                 low_cost_partial_config: Optional[dict] = None,
                 cat_hp_cost: Optional[dict] = None,
                 prune_attr: Optional[str] = None,
                 min_resource: Optional[float] = None,
                 max_resource: Optional[float] = None,
                 reduction_factor: Optional[float] = None,
                 global_search_alg: Optional[Searcher] = None,
                 config_constraints: Optional[
                     List[Tuple[Callable[[dict], float], str, float]]] = None,
                 metric_constraints: Optional[
                     List[Tuple[str, str, float]]] = None,
                 seed: Optional[int] = 20):
        '''Constructor

        Args:
            metric: A string of the metric name to optimize for.
            mode: A string in ['min', 'max'] to specify the objective as
                minimization or maximization.
            space: A dictionary to specify the search space.
            points_to_evaluate: Initial parameter suggestions to be run first.
            low_cost_partial_config: A dictionary from a subset of
                controlled dimensions to the initial low-cost values.
                e.g.,

                .. code-block:: python

                    {'n_estimators': 4, 'max_leaves': 4}

            cat_hp_cost: A dictionary from a subset of categorical dimensions
                to the relative cost of each choice.
                e.g.,

                .. code-block:: python

                    {'tree_method': [1, 1, 2]}

                i.e., the relative cost of the
                three choices of 'tree_method' is 1, 1 and 2 respectively.
            prune_attr: A string of the attribute used for pruning.
                Not necessarily in space.
                When prune_attr is in space, it is a hyperparameter, e.g.,
                    'n_iters', and the best value is unknown.
                When prune_attr is not in space, it is a resource dimension,
                    e.g., 'sample_size', and the peak performance is assumed
                    to be at the max_resource.
            min_resource: A float of the minimal resource to use for the
                prune_attr; only valid if prune_attr is not in space.
            max_resource: A float of the maximal resource to use for the
                prune_attr; only valid if prune_attr is not in space.
            reduction_factor: A float of the reduction factor used for
                incremental pruning.
            global_search_alg: A Searcher instance as the global search
                instance. If omitted, Optuna is used. The following algos have
                known issues when used as global_search_alg:
                - HyperOptSearch raises exception sometimes
                - TuneBOHB has its own scheduler
            config_constraints: A list of config constraints to be satisfied.
                e.g.,

                .. code-block: python

                    config_constraints = [(mem_size, '<=', 1024**3)]

                mem_size is a function which produces a float number for the bytes
                needed for a config.
                It is used to skip configs which do not fit in memory.
            metric_constraints: A list of metric constraints to be satisfied.
                e.g., `['precision', '>=', 0.9]`
            seed: An integer of the random seed.
        '''
        self._metric, self._mode = metric, mode
        init_config = low_cost_partial_config or {}
        if not init_config:
            logger.warning(
                "No low-cost partial config given to the search algorithm. "
                "For cost-frugal search, "
                "consider providing low-cost values for cost-related hps via "
                "'low_cost_partial_config'."
            )
        self._points_to_evaluate = points_to_evaluate or []
        self._config_constraints = config_constraints
        self._metric_constraints = metric_constraints
        if self._metric_constraints:
            # metric modified by lagrange
            metric += self.lagrange
        if global_search_alg is not None:
            self._gs = global_search_alg
        elif getattr(self, '__name__', None) != 'CFO':
            try:
                gs_seed = seed - 10 if (seed - 10) >= 0 else seed - 11 + (1 << 32)
                self._gs = GlobalSearch(space=space, metric=metric, mode=mode, seed=gs_seed)
            except TypeError:
                self._gs = GlobalSearch(space=space, metric=metric, mode=mode)
        else:
            self._gs = None
        self._ls = self.LocalSearch(
            init_config, metric, mode, cat_hp_cost, space, prune_attr,
            min_resource, max_resource, reduction_factor, self.cost_attr, seed)
        self._init_search()

    def set_search_properties(self,
                              metric: Optional[str] = None,
                              mode: Optional[str] = None,
                              config: Optional[Dict] = None) -> bool:
        if not self._ls.space:
            if metric:
                self._metric = metric
                if self._metric_constraints:
                    # metric modified by lagrange
                    metric += self.lagrange
                    # TODO: don't change metric for global search methods that
                    # can handle constraints already
            if mode:
                self._mode = mode
            self._ls.set_search_properties(metric, mode, config)
            if self._gs is not None:
                self._gs.set_search_properties(metric, mode, config)
            self._init_search()
        if 'time_budget_s' in config:
            time_budget_s = config['time_budget_s']
            if time_budget_s is not None:
                self._deadline = time_budget_s + time.time()
                SearchThread.set_eps(time_budget_s)
        if 'metric_target' in config:
            self._metric_target = config.get('metric_target')
        return True

    def _init_search(self):
        '''initialize the search
        '''
        self._metric_target = np.inf * self._ls.metric_op
        self._search_thread_pool = {
            # id: int -> thread: SearchThread
            0: SearchThread(self._ls.mode, self._gs)
        }
        self._thread_count = 1  # total # threads created
        self._init_used = self._ls.init_config is None
        self._trial_proposed_by = {}  # trial_id: str -> thread_id: int
        self._ls_bound_min = self._ls.normalize(self._ls.init_config)
        self._ls_bound_max = self._ls_bound_min.copy()
        self._gs_admissible_min = self._ls_bound_min.copy()
        self._gs_admissible_max = self._ls_bound_max.copy()
        self._result = {}  # config_signature: tuple -> result: Dict
        self._deadline = np.inf
        if self._metric_constraints:
            self._metric_constraint_satisfied = False
            self._metric_constraint_penalty = [
                self.penalty for _ in self._metric_constraints]
        else:
            self._metric_constraint_satisfied = True
            self._metric_constraint_penalty = None

    def save(self, checkpoint_path: str):
        ''' save states to a checkpoint path
        '''
        save_object = self
        with open(checkpoint_path, "wb") as outputFile:
            pickle.dump(save_object, outputFile)

    def restore(self, checkpoint_path: str):
        ''' restore states from checkpoint
        '''
        with open(checkpoint_path, "rb") as inputFile:
            state = pickle.load(inputFile)
        self._metric_target = state._metric_target
        self._search_thread_pool = state._search_thread_pool
        self._thread_count = state._thread_count
        self._init_used = state._init_used
        self._trial_proposed_by = state._trial_proposed_by
        self._ls_bound_min = state._ls_bound_min
        self._ls_bound_max = state._ls_bound_max
        self._gs_admissible_min = state._gs_admissible_min
        self._gs_admissible_max = state._gs_admissible_max
        self._result = state._result
        self._deadline = state._deadline
        self._metric, self._mode = state._metric, state._mode
        self._points_to_evaluate = state._points_to_evaluate
        self._gs = state._gs
        self._ls = state._ls
        self._config_constraints = state._config_constraints
        self._metric_constraints = state._metric_constraints
        self._metric_constraint_satisfied = state._metric_constraint_satisfied
        self._metric_constraint_penalty = state._metric_constraint_penalty

    @property
    def metric_target(self):
        return self._metric_target

    def on_trial_complete(self, trial_id: str, result: Optional[Dict] = None,
                          error: bool = False):
        ''' search thread updater and cleaner
        '''
        metric_constraint_satisfied = True
        if result and not error and self._metric_constraints:
            # account for metric constraints if any
            objective = result[self._metric]
            for i, constraint in enumerate(self._metric_constraints):
                metric_constraint, sign, threshold = constraint
                value = result.get(metric_constraint)
                if value:
                    # sign is <= or >=
                    sign_op = 1 if sign == '<=' else -1
                    violation = (value - threshold) * sign_op
                    if violation > 0:
                        # add penalty term to the metric
                        objective += self._metric_constraint_penalty[
                            i] * violation * self._ls.metric_op
                        metric_constraint_satisfied = False
                        if self._metric_constraint_penalty[i] < self.penalty:
                            self._metric_constraint_penalty[i] += violation
            result[self._metric + self.lagrange] = objective
            if metric_constraint_satisfied and not self._metric_constraint_satisfied:
                # found a feasible point
                self._metric_constraint_penalty = [1 for _ in self._metric_constraints]
            self._metric_constraint_satisfied |= metric_constraint_satisfied
        thread_id = self._trial_proposed_by.get(trial_id)
        if thread_id in self._search_thread_pool:
            self._search_thread_pool[thread_id].on_trial_complete(
                trial_id, result, error)
            del self._trial_proposed_by[trial_id]
        if result:
            config = {}
            for key, value in result.items():
                if key.startswith('config/'):
                    config[key[7:]] = value
            if error:  # remove from result cache
                del self._result[self._ls.config_signature(config)]
            else:  # add to result cache
                self._result[self._ls.config_signature(config)] = result
                # update target metric if improved
                objective = result[
                    self._metric + self.lagrange] if self._metric_constraints \
                    else result[self._metric]
                if (objective - self._metric_target) * self._ls.metric_op < 0:
                    self._metric_target = objective
                if not thread_id and metric_constraint_satisfied \
                        and self._create_condition(result):
                    # thread creator
                    self._search_thread_pool[self._thread_count] = SearchThread(
                        self._ls.mode,
                        self._ls.create(
                            config, objective,
                            cost=result.get(self.cost_attr, 1)),
                        self.cost_attr
                    )
                    thread_id = self._thread_count
                    self._thread_count += 1
                    self._update_admissible_region(
                        config, self._ls_bound_min, self._ls_bound_max)
                elif thread_id and not self._metric_constraint_satisfied:
                    # no point has been found to satisfy metric constraint
                    self._expand_admissible_region()
                # reset admissible region to ls bounding box
                self._gs_admissible_min.update(self._ls_bound_min)
                self._gs_admissible_max.update(self._ls_bound_max)
        # cleaner
        if thread_id and thread_id in self._search_thread_pool:
            # local search thread
            self._clean(thread_id)

    def _update_admissible_region(self, config, admissible_min, admissible_max):
        # update admissible region
        normalized_config = self._ls.normalize(config)
        for key in admissible_min:
            value = normalized_config[key]
            if value > admissible_max[key]:
                admissible_max[key] = value
            elif value < admissible_min[key]:
                admissible_min[key] = value

    def _create_condition(self, result: Dict) -> bool:
        ''' create thread condition
        '''
        if len(self._search_thread_pool) < 2:
            return True
        obj_median = np.median(
            [thread.obj_best1 for id, thread in self._search_thread_pool.items()
             if id])
        return result[self._metric] * self._ls.metric_op < obj_median

    def _clean(self, thread_id: int):
        ''' delete thread and increase admissible region if converged,
        merge local threads if they are close
        '''
        assert thread_id
        todelete = set()
        for id in self._search_thread_pool:
            if id and id != thread_id:
                if self._inferior(id, thread_id):
                    todelete.add(id)
        for id in self._search_thread_pool:
            if id and id != thread_id:
                if self._inferior(thread_id, id):
                    todelete.add(thread_id)
                    break
        if self._search_thread_pool[thread_id].converged:
            todelete.add(thread_id)
            self._expand_admissible_region()
        for id in todelete:
            del self._search_thread_pool[id]

    def _expand_admissible_region(self):
        for key in self._ls_bound_max:
            self._ls_bound_max[key] += self._ls.STEPSIZE
            self._ls_bound_min[key] -= self._ls.STEPSIZE

    def _inferior(self, id1: int, id2: int) -> bool:
        ''' whether thread id1 is inferior to id2
        '''
        t1 = self._search_thread_pool[id1]
        t2 = self._search_thread_pool[id2]
        if t1.obj_best1 < t2.obj_best2:
            return False
        elif t1.resource and t1.resource < t2.resource:
            return False
        elif t2.reach(t1):
            return True
        return False

    def on_trial_result(self, trial_id: str, result: Dict):
        ''' receive intermediate result
        '''
        if trial_id not in self._trial_proposed_by:
            return
        thread_id = self._trial_proposed_by[trial_id]
        if thread_id not in self._search_thread_pool:
            return
        if result and self._metric_constraints:
            result[self._metric + self.lagrange] = result[self._metric]
        self._search_thread_pool[thread_id].on_trial_result(trial_id, result)

    def suggest(self, trial_id: str) -> Optional[Dict]:
        ''' choose thread, suggest a valid config
        '''
        if self._init_used and not self._points_to_evaluate:
            choice, backup = self._select_thread()
            if choice < 0:  # timeout
                return None
            config = self._search_thread_pool[choice].suggest(trial_id)
            if choice and config is None:
                # local search thread finishes
                if self._search_thread_pool[choice].converged:
                    self._expand_admissible_region()
                    del self._search_thread_pool[choice]
                return None
            # preliminary check; not checking config validation
            skip = self._should_skip(choice, trial_id, config)
            if skip:
                if choice:
                    return None
                # use rs when BO fails to suggest a config
                for _, generated in generate_variants({'config': self._ls.space}):
                    config = generated['config']
                    break  # get one random config
                skip = self._should_skip(-1, trial_id, config)
                if skip:
                    return None
            if choice or self._valid(config):
                # LS or valid or no backup choice
                self._trial_proposed_by[trial_id] = choice
            else:  # invalid config proposed by GS
                if choice == backup:
                    # use CFO's init point
                    init_config = self._ls.init_config
                    config = self._ls.complete_config(
                        init_config, self._ls_bound_min, self._ls_bound_max)
                    self._trial_proposed_by[trial_id] = choice
                else:
                    config = self._search_thread_pool[backup].suggest(trial_id)
                    skip = self._should_skip(backup, trial_id, config)
                    if skip:
                        return None
                    self._trial_proposed_by[trial_id] = backup
                    choice = backup
            if not choice:  # global search
                if self._ls._resource:
                    # TODO: min or median?
                    config[self._ls.prune_attr] = self._ls.min_resource
                # temporarily relax admissible region for parallel proposals
                self._update_admissible_region(
                    config, self._gs_admissible_min, self._gs_admissible_max)
            else:
                self._update_admissible_region(
                    config, self._ls_bound_min, self._ls_bound_max)
                self._gs_admissible_min.update(self._ls_bound_min)
                self._gs_admissible_max.update(self._ls_bound_max)
            self._result[self._ls.config_signature(config)] = {}
        else:  # use init config
            init_config = self._points_to_evaluate.pop(
                0) if self._points_to_evaluate else self._ls.init_config
            config = self._ls.complete_config(
                init_config, self._ls_bound_min, self._ls_bound_max)
            config_signature = self._ls.config_signature(config)
            result = self._result.get(config_signature)
            if result:  # tried before
                return None
            elif result is None:  # not tried before
                self._result[config_signature] = {}
            else:  # running but no result yet
                return None
            self._init_used = True
            self._trial_proposed_by[trial_id] = 0
            self._search_thread_pool[0].running += 1
        return config

    def _should_skip(self, choice, trial_id, config) -> bool:
        ''' if config is None or config's result is known or constraints are violated
            return True; o.w. return False
        '''
        if config is None:
            return True
        config_signature = self._ls.config_signature(config)
        exists = config_signature in self._result
        # check constraints
        if not exists and self._config_constraints:
            for constraint in self._config_constraints:
                func, sign, threshold = constraint
                value = func(config)
                if (sign == '<=' and value > threshold
                        or sign == '>=' and value < threshold):
                    self._result[config_signature] = {
                        self._metric: np.inf * self._ls.metric_op,
                        'time_total_s': 1,
                    }
                    exists = True
                    break
        if exists:  # suggested before
            if choice >= 0:  # not fallback to rs
                result = self._result.get(config_signature)
                if result:  # finished
                    self._search_thread_pool[choice].on_trial_complete(
                        trial_id, result, error=False)
                    if choice:
                        # local search thread
                        self._clean(choice)
                # else:     # running
                #     # tell the thread there is an error
                #     self._search_thread_pool[choice].on_trial_complete(
                #         trial_id, {}, error=True)
            return True
        return False

    def _select_thread(self) -> Tuple:
        ''' thread selector; use can_suggest to check LS availability
        '''
        # update priority
        min_eci = self._deadline - time.time()
        if min_eci <= 0:
            # return -1, -1
            # keep proposing new configs assuming no budget left
            min_eci = 0
        max_speed = 0
        for thread in self._search_thread_pool.values():
            if thread.speed > max_speed:
                max_speed = thread.speed
        for thread in self._search_thread_pool.values():
            thread.update_eci(self._metric_target, max_speed)
            if thread.eci < min_eci:
                min_eci = thread.eci
        for thread in self._search_thread_pool.values():
            thread.update_priority(min_eci)

        top_thread_id = backup_thread_id = 0
        priority1 = priority2 = self._search_thread_pool[0].priority
        for thread_id, thread in self._search_thread_pool.items():
            # if thread_id:
            #     print(
            #         f"priority of thread {thread_id}={thread.priority}")
            #     logger.debug(
            #         f"thread {thread_id}.can_suggest={thread.can_suggest}")
            if thread_id and thread.can_suggest:
                priority = thread.priority
                if priority > priority1:
                    priority1 = priority
                    top_thread_id = thread_id
                if priority > priority2 or backup_thread_id == 0:
                    priority2 = priority
                    backup_thread_id = thread_id
        return top_thread_id, backup_thread_id

    def _valid(self, config: Dict) -> bool:
        ''' config validator
        '''
        normalized_config = self._ls.normalize(config)
        for key in self._gs_admissible_min:
            if key in config:
                value = normalized_config[key]
                if value + self._ls.STEPSIZE < self._gs_admissible_min[key] \
                        or value > self._gs_admissible_max[key] + self._ls.STEPSIZE:
                    return False
        return True


try:
    from ray.tune import (uniform, quniform, choice, randint, qrandint, randn,
                          qrandn, loguniform, qloguniform)
except ImportError:
    from ..tune.sample import (uniform, quniform, choice, randint, qrandint, randn,
                               qrandn, loguniform, qloguniform)

try:
    from nni.tuner import Tuner as NNITuner
    from nni.utils import extract_scalar_reward

    class BlendSearchTuner(BlendSearch, NNITuner):
        '''Tuner class for NNI
        '''

        def receive_trial_result(self, parameter_id, parameters, value,
                                 **kwargs):
            '''
            Receive trial's final result.
            parameter_id: int
            parameters: object created by 'generate_parameters()'
            value: final metrics of the trial, including default metric
            '''
            result = {}
            for key, value in parameters.items():
                result['config/' + key] = value
            reward = extract_scalar_reward(value)
            result[self._metric] = reward
            # if nni does not report training cost,
            # using sequence as an approximation.
            # if no sequence, using a constant 1
            result[self.cost_attr] = value.get(self.cost_attr, value.get(
                'sequence', 1))
            self.on_trial_complete(str(parameter_id), result)
        ...

        def generate_parameters(self, parameter_id, **kwargs) -> Dict:
            '''
            Returns a set of trial (hyper-)parameters, as a serializable object
            parameter_id: int
            '''
            return self.suggest(str(parameter_id))
        ...

        def update_search_space(self, search_space):
            '''
            Tuners are advised to support updating search space at run-time.
            If a tuner can only set search space once before generating first hyper-parameters,
            it should explicitly document this behaviour.
            search_space: JSON object created by experiment owner
            '''
            config = {}
            for key, value in search_space.items():
                v = value.get("_value")
                _type = value['_type']
                if _type == 'choice':
                    config[key] = choice(v)
                elif _type == 'randint':
                    config[key] = randint(v[0], v[1] - 1)
                elif _type == 'uniform':
                    config[key] = uniform(v[0], v[1])
                elif _type == 'quniform':
                    config[key] = quniform(v[0], v[1], v[2])
                elif _type == 'loguniform':
                    config[key] = loguniform(v[0], v[1])
                elif _type == 'qloguniform':
                    config[key] = qloguniform(v[0], v[1], v[2])
                elif _type == 'normal':
                    config[key] = randn(v[1], v[2])
                elif _type == 'qnormal':
                    config[key] = qrandn(v[1], v[2], v[3])
                else:
                    raise ValueError(
                        f'unsupported type in search_space {_type}')
            self._ls.set_search_properties(None, None, config)
            if self._gs is not None:
                self._gs.set_search_properties(None, None, config)
            self._init_search()

except ImportError:
    class BlendSearchTuner(BlendSearch):
        pass


class CFO(BlendSearchTuner):
    ''' class for CFO algorithm
    '''

    __name__ = 'CFO'

    def suggest(self, trial_id: str) -> Optional[Dict]:
        # Number of threads is 1 or 2. Thread 0 is a vacuous thread
        assert len(self._search_thread_pool) < 3, len(self._search_thread_pool)
        if len(self._search_thread_pool) < 2:
            # When a local converges, the number of threads is 1
            # Need to restart
            self._init_used = False
        return super().suggest(trial_id)

    def _select_thread(self) -> Tuple:
        for key in self._search_thread_pool:
            if key:
                return key, key

    def _create_condition(self, result: Dict) -> bool:
        ''' create thread condition
        '''
        return len(self._search_thread_pool) < 2



import itertools
from collections import OrderedDict

import ray
from ray.tune.suggest.optuna import OptunaSearch

from flaml import CFO, BlendSearch

SEARCH_ALGO_MAPPING = OrderedDict(
    [
        ("optuna", OptunaSearch),
        ("cfo", CFO),
        ("bs", BlendSearch),
        ("grid", None),
        ("gridbert", None),
        ("rs", None)
    ]
)


class AutoSearchAlgorithm:
    """
    This is a class for getting the search algorithm based on the search algorithm name
    (a string variable) instantiated as one of the algorithms of the library when
    created with the `~flaml.nlp.hpo.AutoSearchAlgorithm.from_method_name` method.

    This class cannot be instantiated directly using ``__init__()`` (throws an error).
    """

    def __init__(self):
        raise EnvironmentError(
            "AutoSearchAlgorithm is designed to be instantiated "
            "using the `AutoSearchAlgorithm.from_method_name(cls, search_algo_name, search_algo_args_mode,"
            " hpo_search_space, **custom_hpo_args)` methods."
        )

    @classmethod
    def from_method_name(cls, search_algo_name, search_algo_args_mode, hpo_search_space, **custom_hpo_args):
        """
        Instantiating one of the search algorithm classes based on the search algorithm name, search algorithm
        argument mode, hpo search space and other keyword args

        Args:
            search_algo_name:
                A string variable that specifies the search algorithm name, e.g., "bs"

            search_algo_args_mode:
                A string variable that specifies the mode for the search algorithm args, e.g., "dft" means
                initializing using the default mode

            hpo_search_space:
                The hpo search space

            custom_hpo_args:
                The customized arguments for the search algorithm (specified by user)

        Example:
        >>> from flaml.nlp.hpo.hpo_searchspace import AutoHPOSearchSpace
        >>> search_space_hpo=AutoHPOSearchSpace.from_model_and_dataset_name("uni", "electra", "small", ["glue"], "rte")
        >>> search_algo = AutoSearchAlgorithm.from_method_name("bs", "cus", search_space_hpo,
                         {"points_to_evaluate": [{"learning_rate": 1e-5, "num_train_epochs": 10}])
        """

        assert hpo_search_space, "hpo_search_space needs to be specified for calling AutoSearchAlgorithm.from_method_name"
        if not search_algo_name:
            search_algo_name = "grid"
        if search_algo_name in SEARCH_ALGO_MAPPING.keys():
            if SEARCH_ALGO_MAPPING[search_algo_name] is None:
                return None
            """
            filtering the customized args for hpo from custom_hpo_args, keep those
            which are in the input variable name list of the constructor of
            the algorithm, remove those which does not appear in the input variables
            of the constructor function
            """
            this_search_algo_kwargs = None
            allowed_arguments = SEARCH_ALGO_MAPPING[search_algo_name].__init__.__code__.co_varnames
            allowed_custom_args = {key: custom_hpo_args[key] for key in custom_hpo_args.keys() if
                                   key in allowed_arguments}

            """
             If the search_algo_args_mode is "dft", set the args to the default args, e.g.,the default args for
             BlendSearch is "low_cost_partial_config": {"num_train_epochs": min_epoch,"per_device_train_batch_size"
             : max(hpo_search_space["per_device_train_batch_size"].categories)},
            """
            if search_algo_args_mode == "dft":
                this_search_algo_kwargs = DEFAULT_SEARCH_ALGO_ARGS_MAPPING[search_algo_name](
                    "dft", hpo_search_space=hpo_search_space, **allowed_custom_args)
            elif search_algo_args_mode == "cus":
                this_search_algo_kwargs = DEFAULT_SEARCH_ALGO_ARGS_MAPPING[search_algo_name](
                    "cus", hpo_search_space=hpo_search_space, **allowed_custom_args)

            """
            returning the hpo algorithm with the arguments
            """
            return SEARCH_ALGO_MAPPING[search_algo_name](**this_search_algo_kwargs)
        raise ValueError(
            "Unrecognized method {} for this kind of AutoSearchAlgorithm: {}.\n"
            "Method name should be one of {}.".format(
                search_algo_name, cls.__name__, ", ".join(SEARCH_ALGO_MAPPING.keys())
            )
        )

    @staticmethod
    def grid2list(grid_config):
        key_val_list = [[(key, each_val) for each_val in val_list['grid_search']]
                        for (key, val_list) in grid_config.items()]
        config_list = [dict(x) for x in itertools.product(*key_val_list)]
        return config_list


def get_search_algo_args_optuna(search_args_mode, hpo_search_space=None, **custom_hpo_args):
    return {}


def default_search_algo_args_bs(search_args_mode, hpo_search_space=None, **custom_hpo_args):
    assert hpo_search_space, "hpo_search_space needs to be specified for calling AutoSearchAlgorithm.from_method_name"
    if "num_train_epochs" in hpo_search_space and \
            isinstance(hpo_search_space["num_train_epochs"], ray.tune.sample.Categorical):
        min_epoch = min(hpo_search_space["num_train_epochs"].categories)
    else:
        assert isinstance(hpo_search_space["num_train_epochs"], ray.tune.sample.Float)
        min_epoch = hpo_search_space["num_train_epochs"].lower
    default_search_algo_args = {
        "low_cost_partial_config": {
            "num_train_epochs": min_epoch,
            "per_device_train_batch_size": max(hpo_search_space["per_device_train_batch_size"].categories),
        },
    }
    if search_args_mode == "cus":
        default_search_algo_args.update(custom_hpo_args)
    return default_search_algo_args


def experiment_search_algo_args_bs(hpo_search_space=None):
    if "num_train_epochs" in hpo_search_space and \
            isinstance(hpo_search_space["num_train_epochs"], ray.tune.sample.Categorical):
        min_epoch = min(hpo_search_space["num_train_epochs"].categories)
    else:
        assert isinstance(hpo_search_space["num_train_epochs"], ray.tune.sample.Float)
        min_epoch = hpo_search_space["num_train_epochs"].lower
    default_search_algo_args = {
        "low_cost_partial_config": {
            "num_train_epochs": min_epoch,
        },
    }
    return default_search_algo_args


def default_search_algo_args_skopt(hpo_search_space=None):
    return {}


def default_search_algo_args_dragonfly(hpo_search_space=None):
    return {}


def default_search_algo_args_nevergrad(hpo_search_space=None):
    return {}


def default_search_algo_args_hyperopt(hpo_search_space=None):
    return {}


def default_search_algo_args_grid_search(search_args_mode, hpo_search_space=None, **custom_hpo_args):
    return {}


def default_search_algo_args_random_search(search_args_mode, hpo_search_space=None, **custom_hpo_args):
    return {}


DEFAULT_SEARCH_ALGO_ARGS_MAPPING = OrderedDict(
    [
        ("optuna", get_search_algo_args_optuna),
        ("cfo", default_search_algo_args_bs),
        ("bs", default_search_algo_args_bs),
        ("grid", default_search_algo_args_grid_search),
        ("gridbert", default_search_algo_args_random_search)
    ]
)



import os

try:
    from transformers import Trainer as TFTrainer
except ImportError:
    TFTrainer = object


class TrainerForAutoTransformers(TFTrainer):

    def evaluate(self,
                 eval_dataset=None):
        """
            Overriding transformers.Trainer.evaluate by saving state with save_state

            Args:
                eval_dataset:
                    the dataset to be evaluated
        """
        from ray import tune

        eval_dataloader = self.get_eval_dataloader(eval_dataset)
        output = self.prediction_loop(
            eval_dataloader, description="Evaluation")
        self.log(output.metrics)

        self.save_state()

        for key in list(output.metrics.keys()):
            if key.startswith("eval_"):
                output.metrics[key[5:]] = output.metrics[key]
        tune.report(**output.metrics)

        return output.metrics

    def save_state(self):
        """
                Overriding transformers.Trainer.save_state. It is only through saving
                the states can best_trial.get_best_checkpoint return a non-empty value.
        """
        import torch
        from transformers.trainer_utils import PREFIX_CHECKPOINT_DIR
        from ray import tune

        with tune.checkpoint_dir(step=self.state.global_step) as checkpoint_dir:
            self.args.output_dir = checkpoint_dir
            # This is the directory name that Huggingface requires.
            output_dir = os.path.join(
                self.args.output_dir,
                f"{PREFIX_CHECKPOINT_DIR}-{self.state.global_step}")
            self.save_model(output_dir)
            torch.save(self.optimizer.state_dict(),
                       os.path.join(output_dir, "optimizer.pt"))
            torch.save(self.lr_scheduler.state_dict(),
                       os.path.join(output_dir, "scheduler.pt"))

    @staticmethod
    def convert_num_train_epochs_to_max_steps(
            num_train_epochs: int,
            num_train_examples: int,
            per_device_train_batch_size: int,
            device_count: int):
        return int(num_train_epochs * num_train_examples / per_device_train_batch_size / device_count)

    @staticmethod
    def convert_max_steps_to_num_train_epochs(
            max_steps: int,
            num_train_examples: int,
            per_device_train_batch_size: int,
            device_count: int):
        return float(max_steps * per_device_train_batch_size * device_count) / num_train_examples

    @staticmethod
    def convert_warmup_ratio_to_warmup_steps(
            warmup_ratio,
            max_steps=None,
            num_train_epochs=None,
            num_train_examples=None,
            per_device_train_batch_size=None,
            device_count=None):
        if max_steps:
            return int(warmup_ratio * max_steps)
        max_steps = TrainerForAutoTransformers.convert_num_train_epochs_to_max_steps(
            num_train_epochs,
            num_train_examples,
            per_device_train_batch_size,
            device_count)
        return int(warmup_ratio * max_steps)

    @staticmethod
    def convert_warmup_steps_to_warmup_ratio(
            warmup_steps: int,
            num_train_epochs: int,
            num_train_examples: int,
            per_device_train_batch_size: int,
            device_count: int):
        max_steps = TrainerForAutoTransformers.convert_num_train_epochs_to_max_steps(
            num_train_epochs,
            num_train_examples,
            per_device_train_batch_size,
            device_count)
        return float(warmup_steps / max_steps)



'''
Copyright 2020 The Ray Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This source file is adapted here because ray does not fully support Windows.

Copyright (c) Microsoft Corporation.
'''
import os

# yapf: disable
# __sphinx_doc_begin__
# (Optional/Auto-filled) training is terminated. Filled only if not provided.
DONE = "done"

# (Optional) Enum for user controlled checkpoint
SHOULD_CHECKPOINT = "should_checkpoint"

# (Auto-filled) The hostname of the machine hosting the training process.
HOSTNAME = "hostname"

# (Auto-filled) The auto-assigned id of the trial.
TRIAL_ID = "trial_id"

# (Auto-filled) The auto-assigned id of the trial.
EXPERIMENT_TAG = "experiment_tag"

# (Auto-filled) The node ip of the machine hosting the training process.
NODE_IP = "node_ip"

# (Auto-filled) The pid of the training process.
PID = "pid"

# (Optional) Default (anonymous) metric when using tune.report(x)
DEFAULT_METRIC = "_metric"

# (Optional) Mean reward for current training iteration
EPISODE_REWARD_MEAN = "episode_reward_mean"

# (Optional) Mean loss for training iteration
MEAN_LOSS = "mean_loss"

# (Optional) Mean loss for training iteration
NEG_MEAN_LOSS = "neg_mean_loss"

# (Optional) Mean accuracy for training iteration
MEAN_ACCURACY = "mean_accuracy"

# Number of episodes in this iteration.
EPISODES_THIS_ITER = "episodes_this_iter"

# (Optional/Auto-filled) Accumulated number of episodes for this trial.
EPISODES_TOTAL = "episodes_total"

# Number of timesteps in this iteration.
TIMESTEPS_THIS_ITER = "timesteps_this_iter"

# (Auto-filled) Accumulated number of timesteps for this entire trial.
TIMESTEPS_TOTAL = "timesteps_total"

# (Auto-filled) Time in seconds this iteration took to run.
# This may be overridden to override the system-computed time difference.
TIME_THIS_ITER_S = "time_this_iter_s"

# (Auto-filled) Accumulated time in seconds for this entire trial.
TIME_TOTAL_S = "time_total_s"

# (Auto-filled) The index of this training iteration.
TRAINING_ITERATION = "training_iteration"
# __sphinx_doc_end__
# yapf: enable

DEFAULT_EXPERIMENT_INFO_KEYS = ("trainable_name", EXPERIMENT_TAG, TRIAL_ID)

DEFAULT_RESULT_KEYS = (TRAINING_ITERATION, TIME_TOTAL_S, TIMESTEPS_TOTAL,
                       MEAN_ACCURACY, MEAN_LOSS)

# Make sure this doesn't regress
AUTO_RESULT_KEYS = (
    TRAINING_ITERATION,
    TIME_TOTAL_S,
    EPISODES_TOTAL,
    TIMESTEPS_TOTAL,
    NODE_IP,
    HOSTNAME,
    PID,
    TIME_TOTAL_S,
    TIME_THIS_ITER_S,
    "timestamp",
    "experiment_id",
    "date",
    "time_since_restore",
    "iterations_since_restore",
    "timesteps_since_restore",
    "config",
)

# __duplicate__ is a magic keyword used internally to
# avoid double-logging results when using the Function API.
RESULT_DUPLICATE = "__duplicate__"

# __trial_info__ is a magic keyword used internally to pass trial_info
# to the Trainable via the constructor.
TRIAL_INFO = "__trial_info__"

# __stdout_file__/__stderr_file__ are magic keywords used internally
# to pass log file locations to the Trainable via the constructor.
STDOUT_FILE = "__stdout_file__"
STDERR_FILE = "__stderr_file__"

# Where Tune writes result files by default
DEFAULT_RESULTS_DIR = (os.environ.get("TEST_TMPDIR")
                       or os.environ.get("TUNE_RESULT_DIR")
                       or os.path.expanduser("~/ray_results"))

# Meta file about status under each experiment directory, can be
# parsed by automlboard if exists.
JOB_META_FILE = "job_status.json"

# Meta file about status under each trial directory, can be parsed
# by automlboard if exists.
EXPR_META_FILE = "trial_status.json"

# File that stores parameters of the trial.
EXPR_PARAM_FILE = "params.json"

# Pickle File that stores parameters of the trial.
EXPR_PARAM_PICKLE_FILE = "params.pkl"

# File that stores the progress of the trial.
EXPR_PROGRESS_FILE = "progress.csv"

# File that stores results of the trial.
EXPR_RESULT_FILE = "result.json"

# Config prefix when using Analysis.
CONFIG_PREFIX = "config/"



from collections import OrderedDict

from .get_grid_search_space import \
    (get_electra_space,
     get_bert_space,
     get_roberta_space,
     get_funnel_space,
     get_deberta_space,
     get_albert_space,
     get_longformer_space,
     get_mobilebert_space
     )

GRID_SEARCH_SPACE_MAPPING = OrderedDict(
    [
        ("electra", get_electra_space),
        ("bert", get_bert_space),
        ("roberta", get_roberta_space),
        ("funnel", get_funnel_space),
        ("deberta", get_deberta_space),
        ("albert", get_albert_space),
        ("mobilebert", get_mobilebert_space),
        ("longformer", get_longformer_space)
    ]
)

HF_MODEL_LIST = [
    "bert",
    "roberta",
    "electra",
    "xlnet",
    "albert",
    "distilbert",
    "deberta",
    "mobilebert",
    "funnel"
]


class AutoGridSearchSpace:
    """
    This is a class for getting the recommended grid search space of a pre-trained LM that will be
    instantiated as one of the search spaces of the library when created with the
    `~flaml.nlp.hpo.AutoGridSearchSpace.from_model_and_dataset_name` method.

    This class cannot be instantiated directly using ``__init__()`` (throws an error).
    """

    def __init__(self):
        raise EnvironmentError(
            "AutoGridSearchSpace is designed to be instantiated "
            "using the `AutoGridSearchSpace.from_config_and_method_name(cls, model_type, model_size_type,"
            "dataset_name,subdataset_name = None,algo_mode = None)` methods."
        )

    @classmethod
    def from_model_and_dataset_name(cls,
                                    model_type,
                                    model_size_type,
                                    dataset_name_list: list = None,
                                    subdataset_name=None,
                                    algo_mode=None):
        """
        Instantiate one of the classes for getting the recommended grid search space of a pre-trained LM from
        the model type, model size type, dataset name, sub dataset name and algorithm mode

        Args:
            model_type:
                A string variable which is the model type, e.g. "electra"

            model_size_type:
                A string variable which is the size of the model, e.g., "small"

            dataset_name_list:
                A string variable which is the dataset name, e.g., "glue"

            subdataset_name:
                A string variable which is the sub dataset name,e.g., "rte"

            algo_mode:
                A string variable which is the algorithm mode for grid search, e.g., "gridbert"

        Example:
            >>> AutoGridSearchSpace.from_model_and_dataset_name("electra", "small", ["glue"], "rte", "grid")

        """

        if model_type in GRID_SEARCH_SPACE_MAPPING.keys():
            this_model_recommended_space = GRID_SEARCH_SPACE_MAPPING[model_type](
                model_size_type, dataset_name_list, subdataset_name, algo_mode)
            return this_model_recommended_space
        raise ValueError(
            "Unrecognized method {},{} for this kind of AutoGridSearchSpace: {}.\n"
            "Method name should be one of {}.".format(
                model_type, dataset_name_list, cls.__name__, ", ".join(GRID_SEARCH_SPACE_MAPPING.keys())
            )
        )



'''!
 * Copyright (c) 2020-2021 Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
'''

import numpy as np
from scipy.sparse import vstack, issparse
import pandas as pd
from .training_log import training_log_reader

from datetime import datetime


def load_openml_dataset(dataset_id, data_dir=None, random_state=0):
    '''Load dataset from open ML.

    If the file is not cached locally, download it from open ML.

    Args:
        dataset_id: An integer of the dataset id in openml
        data_dir: A string of the path to store and load the data
        random_state: An integer of the random seed for splitting data

    Returns:
        X_train: A dataframe of training data
        X_test:  A dataframe of test data
        y_train: A series of labels for training data
        y_test:  A series of labels for test data
    '''
    import os
    import openml
    import pickle
    from sklearn.model_selection import train_test_split

    filename = 'openml_ds' + str(dataset_id) + '.pkl'
    filepath = os.path.join(data_dir, filename)
    if os.path.isfile(filepath):
        print('load dataset from', filepath)
        with open(filepath, 'rb') as f:
            dataset = pickle.load(f)
    else:
        print('download dataset from openml')
        dataset = openml.datasets.get_dataset(dataset_id)
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)
        with open(filepath, 'wb') as f:
            pickle.dump(dataset, f, pickle.HIGHEST_PROTOCOL)
    print('Dataset name:', dataset.name)
    X, y, * \
        __ = dataset.get_data(
            target=dataset.default_target_attribute)
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, random_state=random_state)
    print(
        'X_train.shape: {}, y_train.shape: {};\nX_test.shape: {}, y_test.shape: {}'.format(
            X_train.shape, y_train.shape, X_test.shape, y_test.shape,
        )
    )
    return X_train, X_test, y_train, y_test


def load_openml_task(task_id, data_dir):
    '''Load task from open ML.

    Use the first fold of the task.
    If the file is not cached locally, download it from open ML.

    Args:
        task_id: An integer of the task id in openml
        data_dir: A string of the path to store and load the data

    Returns:
        X_train: A dataframe of training data
        X_test:  A dataframe of test data
        y_train: A series of labels for training data
        y_test:  A series of labels for test data
    '''
    import os
    import openml
    import pickle
    task = openml.tasks.get_task(task_id)
    filename = 'openml_task' + str(task_id) + '.pkl'
    filepath = os.path.join(data_dir, filename)
    if os.path.isfile(filepath):
        print('load dataset from', filepath)
        with open(filepath, 'rb') as f:
            dataset = pickle.load(f)
    else:
        print('download dataset from openml')
        dataset = task.get_dataset()
        with open(filepath, 'wb') as f:
            pickle.dump(dataset, f, pickle.HIGHEST_PROTOCOL)
    X, y, _, _ = dataset.get_data(task.target_name)
    train_indices, test_indices = task.get_train_test_split_indices(
        repeat=0,
        fold=0,
        sample=0,
    )
    X_train = X.iloc[train_indices]
    y_train = y[train_indices]
    X_test = X.iloc[test_indices]
    y_test = y[test_indices]
    print(
        'X_train.shape: {}, y_train.shape: {},\nX_test.shape: {}, y_test.shape: {}'.format(
            X_train.shape, y_train.shape, X_test.shape, y_test.shape,
        )
    )
    return X_train, X_test, y_train, y_test


def get_output_from_log(filename, time_budget):
    '''Get output from log file

    Args:
        filename: A string of the log file name
        time_budget: A float of the time budget in seconds

    Returns:
        training_time_list: A list of the finished time of each logged iter
        best_error_list:
            A list of the best validation error after each logged iter
        error_list: A list of the validation error of each logged iter
        config_list:
            A list of the estimator, sample size and config of each logged iter
        logged_metric_list: A list of the logged metric of each logged iter
    '''

    best_config = None
    best_learner = None
    best_val_loss = float('+inf')
    training_duration = 0.0

    training_time_list = []
    config_list = []
    best_error_list = []
    error_list = []
    logged_metric_list = []
    best_config_list = []
    with training_log_reader(filename) as reader:
        for record in reader.records():
            time_used = record.total_search_time
            training_duration = time_used
            val_loss = record.validation_loss
            config = record.config
            learner = record.learner.split('_')[0]
            sample_size = record.sample_size
            train_loss = record.logged_metric

            if time_used < time_budget:
                if val_loss < best_val_loss:
                    best_val_loss = val_loss
                    best_config = config
                    best_learner = learner
                    best_config_list.append(best_config)
                training_time_list.append(training_duration)
                best_error_list.append(best_val_loss)
                logged_metric_list.append(train_loss)
                error_list.append(val_loss)
                config_list.append({"Current Learner": learner,
                                    "Current Sample": sample_size,
                                    "Current Hyper-parameters": record.config,
                                    "Best Learner": best_learner,
                                    "Best Hyper-parameters": best_config})

    return (training_time_list, best_error_list, error_list, config_list,
            logged_metric_list)


def concat(X1, X2):
    '''concatenate two matrices vertically
    '''
    if isinstance(X1, pd.DataFrame) or isinstance(X1, pd.Series):
        df = pd.concat([X1, X2], sort=False)
        df.reset_index(drop=True, inplace=True)
        if isinstance(X1, pd.DataFrame):
            cat_columns = X1.select_dtypes(
                include='category').columns
            if len(cat_columns):
                df[cat_columns] = df[cat_columns].astype('category')
        return df
    if issparse(X1):
        return vstack((X1, X2))
    else:
        return np.concatenate([X1, X2])


class DataTransformer:
    '''transform X, y
    '''

    def fit_transform(self, X, y, task):
        if isinstance(X, pd.DataFrame):
            X = X.copy()
            n = X.shape[0]
            cat_columns, num_columns, datetime_columns = [], [], []
            drop = False
            for column in X.columns:
                # sklearn\utils\validation.py needs int/float values
                if X[column].dtype.name in ('object', 'category'):
                    if X[column].nunique() == 1 or X[column].nunique(
                            dropna=True) == n - X[column].isnull().sum():
                        X.drop(columns=column, inplace=True)
                        drop = True
                    elif X[column].dtype.name == 'category':
                        current_categories = X[column].cat.categories
                        if '__NAN__' not in current_categories:
                            X[column] = X[column].cat.add_categories(
                                '__NAN__').fillna('__NAN__')
                        cat_columns.append(column)
                    else:
                        X[column] = X[column].fillna('__NAN__')
                        cat_columns.append(column)
                else:
                    # print(X[column].dtype.name)
                    if X[column].nunique(dropna=True) < 2:
                        X.drop(columns=column, inplace=True)
                        drop = True
                    else:
                        if X[column].dtype.name == 'datetime64[ns]':
                            tmp_dt = X[column].dt
                            new_columns_dict = {f'year_{column}': tmp_dt.year, f'month_{column}': tmp_dt.month,
                                                f'day_{column}': tmp_dt.day, f'hour_{column}': tmp_dt.hour,
                                                f'minute_{column}': tmp_dt.minute, f'second_{column}': tmp_dt.second,
                                                f'dayofweek_{column}': tmp_dt.dayofweek,
                                                f'dayofyear_{column}': tmp_dt.dayofyear,
                                                f'quarter_{column}': tmp_dt.quarter}
                            for new_col_name in new_columns_dict.keys():
                                if new_col_name not in X.columns and \
                                        new_columns_dict.get(new_col_name).nunique(dropna=False) >= 2:
                                    X[new_col_name] = new_columns_dict.get(new_col_name)
                                    num_columns.append(new_col_name)
                            X[column] = X[column].map(datetime.toordinal)
                            datetime_columns.append(column)
                            del tmp_dt
                        else:
                            X[column] = X[column].fillna(np.nan)
                            num_columns.append(column)
            X = X[cat_columns + num_columns]
            if cat_columns:
                X[cat_columns] = X[cat_columns].astype('category')
            if num_columns:
                X_num = X[num_columns]
                if drop and np.issubdtype(X_num.columns.dtype, np.integer):
                    X_num.columns = range(X_num.shape[1])
                else:
                    drop = False
                from sklearn.impute import SimpleImputer
                from sklearn.compose import ColumnTransformer
                self.transformer = ColumnTransformer([(
                    'continuous',
                    SimpleImputer(missing_values=np.nan, strategy='median'),
                    X_num.columns)])
                X[num_columns] = self.transformer.fit_transform(X_num)
            self._cat_columns, self._num_columns, self._datetime_columns = \
                cat_columns, num_columns, datetime_columns
            self._drop = drop

        if task == 'regression':
            self.label_transformer = None
        else:
            from sklearn.preprocessing import LabelEncoder
            self.label_transformer = LabelEncoder()
            y = self.label_transformer.fit_transform(y)
        return X, y

    def transform(self, X):
        X = X.copy()
        if isinstance(X, pd.DataFrame):
            cat_columns, num_columns, datetime_columns = self._cat_columns, \
                self._num_columns, self._datetime_columns
            if datetime_columns:
                for column in datetime_columns:
                    tmp_dt = X[column].dt
                    new_columns_dict = {f'year_{column}': tmp_dt.year, f'month_{column}': tmp_dt.month,
                                        f'day_{column}': tmp_dt.day, f'hour_{column}': tmp_dt.hour,
                                        f'minute_{column}': tmp_dt.minute, f'second_{column}': tmp_dt.second,
                                        f'dayofweek_{column}': tmp_dt.dayofweek,
                                        f'dayofyear_{column}': tmp_dt.dayofyear,
                                        f'quarter_{column}': tmp_dt.quarter}
                    for new_col_name in new_columns_dict.keys():
                        if new_col_name not in X.columns and \
                                new_columns_dict.get(new_col_name).nunique(dropna=False) >= 2:
                            X[new_col_name] = new_columns_dict.get(new_col_name)
                    X[column] = X[column].map(datetime.toordinal)
                    del tmp_dt
            X = X[cat_columns + num_columns].copy()
            for column in cat_columns:
                if X[column].dtype.name == 'object':
                    X[column] = X[column].fillna('__NAN__')
                elif X[column].dtype.name == 'category':
                    current_categories = X[column].cat.categories
                    if '__NAN__' not in current_categories:
                        X[column] = X[column].cat.add_categories(
                            '__NAN__').fillna('__NAN__')
            if cat_columns:
                X[cat_columns] = X[cat_columns].astype('category')
            if num_columns:
                X_num = X[num_columns].fillna(np.nan)
                if self._drop:
                    X_num.columns = range(X_num.shape[1])
                X[num_columns] = self.transformer.transform(X_num)
        return X



# lookup table for the grid configs in each pre-trained language huggingface for different tasks


def get_space_union_and_unique(search_space_common,
                               search_space_unique,
                               this_case_tags: list):
    """
        get the recommended search configs for each pre-trained language models

        Args:
            search_space_common:
                the union of configs recommended by the LM for all cases;
            search_space_unique:
                the recommended config by the LM for a specific condition, e.g., small model
            this_case_tags:
                a list, which contains the tags describing the specific condition, e.g., ["small"]
    """
    search_space_union = search_space_common.copy()
    this_search_space = search_space_common.copy()
    # enumerate over each case where the search space is different
    # this difference can be the dataset or model size, etc.
    is_included = False
    from ..utils import merge_dicts
    for each_case in search_space_unique.keys():
        from ..utils import _check_dict_keys_overlaps
        if each_case in this_case_tags:
            is_included = True
            assert not _check_dict_keys_overlaps(this_search_space, search_space_unique[each_case]), \
                "the hyperparameters of common and unique search spaces should not have overlaps"
            this_search_space.update(search_space_unique[each_case])
        search_space_union = merge_dicts(search_space_union, search_space_unique[each_case])
    if is_included:
        return this_search_space
    else:
        if "other" in search_space_unique.keys():
            search_space_union = merge_dicts(search_space_union, search_space_unique["other"])
        return search_space_union


def get_deberta_space(model_size_type=None,
                      dataset_name_list: list = None,
                      subdataset_name=None,
                      algo_mode=None):
    """
        DEBERTA: DECODING-ENHANCED BERT WITH DISENTANGLED ATTENTION: Table 9
        https://arxiv.org/abs/2006.03654
    """
    search_space_common = {
        "cls_dropout": [0, 0.1, 0.15],
        "warmup_steps": [50, 100, 500, 1000],
        "per_device_train_batch_size": [16, 32, 48, 64],
        "num_train_epochs": [10],
        "adam_epsilon": [1e-6],
    }
    search_space_unique = {
        "large": {
            "learning_rate": [5e-6, 8e-6, 9e-6, 1e-5],
            "weight_decay": [0.01],
        },
        "base": {
            "learning_rate": [1.5e-5, 2e-5, 3e-5, 4e-5],
        }
    }
    return get_space_union_and_unique(search_space_common, search_space_unique, [model_size_type])


def get_longformer_space(model_size_type=None,
                         dataset_name_list: list = None,
                         subdataset_name=None,
                         algo_mode=None):
    """
        TODO: Longformer: The Long-Document Transformer
    """
    return


def get_funnel_space(model_size_type=None,
                     dataset_name_list: list = None,
                     subdataset_name=None,
                     algo_mode=None):
    """
    Funnel-Transformer: Filtering out Sequential Redundancy for Efficient Language Processing
    https://arxiv.org/abs/2006.03236
    """
    search_space_common = {"learning_rate": [1e-5, 2e-5, 3e-5],
                           "hidden_dropout": [0.1],
                           "activation_dropout": [0.0],
                           "attention_dropout": [0.1],
                           "weight_decay": [0.01],
                           "warmup_ratio": [0.1],
                           "adam_epsilon": [1e-6],
                           }
    search_space_unique = {
        "imdb": {
            "per_device_train_batch_size": [32],
            "num_train_epochs": [5]
        },
        "ag_news": {
            "per_device_train_batch_size": [32],
            "num_train_epochs": [3]
        },
        "dbpedia_14": {
            "per_device_train_batch_size": [64],
            "num_train_epochs": [3]
        },
        "yelp_polarity": {
            "per_device_train_batch_size": [128],
            "num_train_epochs": [3]
        },
        "yelp_review_full": {
            "per_device_train_batch_size": [128],
            "num_train_epochs": [3]
        },
        "amazon_polarity": {
            "per_device_train_batch_size": [128],
            "num_train_epochs": [3]
        },
        "amazon_review_multi": {
            "per_device_train_batch_size": [128],
            "num_train_epochs": [3]
        },
        "glue_rte": {
            "per_device_train_batch_size": [16],
            "num_train_epochs": [10]
        },
        "glue_mrpc": {
            "per_device_train_batch_size": [16],
            "num_train_epochs": [10]
        },
        "glue_stsb": {
            "per_device_train_batch_size": [16],
            "num_train_epochs": [10]
        },
        "glue_cola": {
            "per_device_train_batch_size": [16],
            "num_train_epochs": [10]
        },
        "glue_sst2": {
            "per_device_train_batch_size": [32],
            "num_train_epochs": [5]
        },
        "glue_qnli": {
            "per_device_train_batch_size": [32],
            "num_train_epochs": [3]
        },
        "glue_mnli": {
            "per_device_train_batch_size": [64],
            "num_train_epochs": [3]
        },
        "glue_qqp": {
            "per_device_train_batch_size": [64],
            "num_train_epochs": [5]
        }
    }
    from ..result_analysis.azure_utils import JobID
    return get_space_union_and_unique(search_space_common, search_space_unique,
                                      [JobID.get_full_data_name(
                                          dataset_name_list,
                                          subdataset_name)])


def get_bert_space(model_size_type=None,
                   dataset_name_list: list = None,
                   subdataset_name=None,
                   algo_mode=None):
    """
        BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding
        https://arxiv.org/pdf/1810.04805.pdf
    """
    search_space_common = {}
    search_space_unique = {
        # Section 4.1: We use a batch size of 32 and fine-tune for 3 epochs over the data for all GLUE tasks. For each
        # task, we selected the best fine-tuning learning rate (among 5e-5, 4e-5, 3e-5, and 2e-5) on the Dev set
        "glue": {
            "learning_rate": [5e-5, 4e-5, 3e-5, 2e-5],
            "per_device_train_batch_size": [32],
            "num_train_epochs": [3],
        },
        # Section 4.2: We fine-tune for 3 epochs with a learning rate of 5e-5 and a batch size of 32
        "squad": {
            "learning_rate": [5e-5],
            "per_device_train_batch_size": [32],
            "num_train_epochs": [2],
        },
        # Section 4.3: We fine-tuned for 2 epochs with a learning rate of 5e-5 and a batch size of 48.
        "squad_v2": {
            "learning_rate": [5e-5],
            "per_device_train_batch_size": [48],
            "num_train_epochs": [2],
        },
        # Section 4.4: We fine-tune the huggingface for 3 epochs with a learning rate of 2e-5 and a batch size of 16.
        "swag": {
            "learning_rate": [2e-5],
            "per_device_train_batch_size": [16],
            "num_train_epochs": [3],
        },
        # Appedix A. The optimal hyperparameter values are task-specific, but we found the following
        # range of possible values to work well across all tasks:
        # - Batch size: 16, 32
        # - Learning rate (Adam): 5e-5, 3e-5, 2e-5
        # - Number of epochs: 2, 3, 4
        "other": {
            "learning_rate": [5e-5, 3e-5, 2e-5],
            "per_device_train_batch_size": [16, 32],
            "num_train_epochs": [2, 3, 4],
        }
    }
    return get_space_union_and_unique(search_space_common,
                                      search_space_unique,
                                      dataset_name_list)


def get_roberta_space(model_size_type=None,
                      dataset_name_list: list = None,
                      subdataset_name=None,
                      algo_mode=None):
    # RoBERTa: A Robustly Optimized BERT Pretraining Approach
    # https://arxiv.org/pdf/1907.11692.pdf
    search_space_common = {
        "warmup_ratio": [0.06],
    }
    search_space_unique = {
        # Table 10: Hyperparameters for finetuning RoBERTa-LARGE on RACE, SQuAD and GLUE.
        # We consider a limited hyperparameter
        # sweep for each task, with batch sizes ∈ {16, 32}
        # and learning rates ∈ {1e−5, 2e−5, 3e−5}, with a
        # linear warmup for the first 6% of steps followed by
        # a linear decay to 0.
        "glue": {
            "learning_rate": [1e-5, 2e-5, 3e-5],
            "per_device_train_batch_size": [16, 32],
            "weight_decay": [0.1],
            "num_train_epochs": [10],
        },
        "race": {
            "learning_rate": [1e-5],
            "per_device_train_batch_size": [16],
            "weight_decay": [0.1],
            "num_train_epochs": [4],
        },
        "squad": {
            "learning_rate": [1.5e-5],
            "per_device_train_batch_size": [48],
            "weight_decay": [0.01],
            "num_train_epochs": [2],
        }
    }
    return get_space_union_and_unique(search_space_common,
                                      search_space_unique,
                                      dataset_name_list)


def get_electra_space(model_size_type=None,
                      dataset_name_list: list = None,
                      subdataset_name=None,
                      algo_mode=None):
    """
        ELECTRA: PRE-TRAINING TEXT ENCODERS AS DISCRIMINATORS RATHER THAN GENERATORS
        https://arxiv.org/pdf/2003.10555.pdf
    """
    assert model_size_type in ("small", "base", "large", "intermediate", "xlarge"), \
        "Electra paper has only provided hyperparameter for the small and base huggingface"
    search_space_common = {
        "learning_rate": [3e-5, 5e-5, 1e-4, 1.5e-4],
        "weight_decay": [0.0],
        "adam_epsilon": [1e-6],
        "warmup_ratio": [0.1],
        "per_device_train_batch_size": [32],
        "hidden_dropout_prob": [0.1],
        "attention_probs_dropout_prob": [0.1],
    }
    search_space_unique = {
        # Appendix B: For Basesized models we searched for a learning
        "squad": {
            "num_train_epochs": [2]
        },
        "squad_v2": {
            "num_train_epochs": [2]
        },
        "glue_stsb": {
            "num_train_epochs": [10],
        },
        "glue_rte": {
            "num_train_epochs": [10],
        },
        "glue_wnli": {
            "num_train_epochs": [3],
        },
        "glue_mrpc": {
            "num_train_epochs": [3],
        },
        "glue_cola": {
            "num_train_epochs": [3],
        },
        "glue_sst2": {
            "num_train_epochs": [3],
        },
        "glue_qnli": {
            "num_train_epochs": [3],
        },
        "glue_mnli": {
            "num_train_epochs": [3],
        },
        "glue_qqp": {
            "num_train_epochs": [3],
        }
    }
    from ..result_analysis.azure_utils import JobID
    return get_space_union_and_unique(search_space_common, search_space_unique,
                                      [JobID.get_full_data_name(
                                          dataset_name_list,
                                          subdataset_name), model_size_type])


def get_mobilebert_space(model_size_type=None,
                         dataset_name_list: list = None,
                         subdataset_name=None,
                         algo_mode=None):
    """
        MobileBERT: a Compact Task-Agnostic BERT for Resource-Limited Devices
        https://arxiv.org/pdf/2004.02984.pdf
    """
    # To finetune the pre-trained models, we search the optimization hyperparameters
    # in a search space including different batch sizes (16/32/48), learning
    # rates ((1-10) * e-5), and the number of epochs (2-10)
    search_space_common = {
        "learning_rate": [x * 1e-5 for x in range(1, 11)],
        "per_device_train_batch_size": [4, 8, 16, 32, 48],
        "num_train_epochs": [x for x in range(2, 11)],
    }
    search_space_unique = {}
    return get_space_union_and_unique(search_space_common, search_space_unique, [])


def get_albert_space(model_size_type=None,
                     dataset_name_list: list = None,
                     subdataset_name=None,
                     algo_mode=None):
    """
        Hyperparameters for downstream tasks are shown in Table 14. We adapt these hyperparameters
        from Liu et al. (2019), Devlin et al. (2019), and Yang et al. (2019).

        LR BSZ ALBERT DR Classifier DR TS WS MSL
        CoLA 1.00E-05 16 0 0.1 5336 320 512
        STS 2.00E-05 16 0 0.1 3598 214 512
        SST-2 1.00E-05 32 0 0.1 20935 1256 512
        MNLI 3.00E-05 128 0 0.1 10000 1000 512
        QNLI 1.00E-05 32 0 0.1 33112 1986 512
        QQP 5.00E-05 128 0.1 0.1 14000 1000 512
        RTE 3.00E-05 32 0.1 0.1 800 200 512
        MRPC 2.00E-05 32 0 0.1 800 200 512
        WNLI 2.00E-05 16 0.1 0.1 2000 250 512
        SQuAD v1.1 5.00E-05 48 0 0.1 3649 365 384
        SQuAD v2.0 3.00E-05 48 0 0.1 8144 814 512
        RACE 2.00E-05 32 0.1 0.1 12000 1000 512
    """
    search_space_common = {
    }
    search_space_unique = {
        "glue_cola": {
            "learning_rate": [1e-5],
            "per_device_train_batch_size": [16],
            "attention_probs_dropout_prob": [0],
            "classifier_dropout_prob": [0.1],
            "max_steps": [5336],
            "warmup_steps": [320],
        },
        "glue_stsb": {
            "learning_rate": [2e-5],
            "per_device_train_batch_size": [16],
            "attention_probs_dropout_prob": [0],
            "classifier_dropout_prob": [0.1],
            "max_steps": [3598],
            "warmup_steps": [214],
        },
        "glue_sst2": {
            "learning_rate": [1e-5],
            "per_device_train_batch_size": [32],
            "attention_probs_dropout_prob": [0],
            "classifier_dropout_prob": [0.1],
            "max_steps": [20935],
            "warmup_steps": [1256],
        },
        "glue_mnli": {
            "learning_rate": [3e-5],
            "per_device_train_batch_size": [128],
            "attention_probs_dropout_prob": [0],
            "classifier_dropout_prob": [0.1],
            "max_steps": [10000],
            "warmup_steps": [1000],
        },
        "glue_qnli": {
            "learning_rate": [1e-5],
            "per_device_train_batch_size": [32],
            "attention_probs_dropout_prob": [0],
            "classifier_dropout_prob": [0.1],
            "max_steps": [33112],
            "warmup_steps": [1986],
        },
        "glue_qqp": {
            "learning_rate": [5e-5],
            "per_device_train_batch_size": [128],
            "attention_probs_dropout_prob": [0.1],
            "classifier_dropout_prob": [0.1],
            "max_steps": [14000],
            "warmup_steps": [1000],
        },
        "glue_rte": {
            "learning_rate": [3e-5],
            "per_device_train_batch_size": [32],
            "attention_probs_dropout_prob": [0.1],
            "classifier_dropout_prob": [0.1],
            "max_steps": [800],
            "warmup_steps": [200],
        },
        "glue_mrpc": {
            "learning_rate": [2e-5],
            "per_device_train_batch_size": [32],
            "attention_probs_dropout_prob": [0],
            "classifier_dropout_prob": [0.1],
            "max_steps": [800],
            "warmup_steps": [200],
        },
        "glue_wnli": {
            "learning_rate": [2e-5],
            "per_device_train_batch_size": [16],
            "attention_probs_dropout_prob": [0.1],
            "classifier_dropout_prob": [0.1],
            "max_steps": [2000],
            "warmup_steps": [250],
        },
        "squad": {
            "learning_rate": [5e-5],
            "per_device_train_batch_size": [48],
            "attention_probs_dropout_prob": [0],
            "classifier_dropout_prob": [0.1],
            "max_steps": [3649],
            "warmup_steps": [365],
        },
        "squad_v2": {
            "learning_rate": [3e-5],
            "per_device_train_batch_size": [48],
            "attention_probs_dropout_prob": [0],
            "classifier_dropout_prob": [0.1],
            "max_steps": [8144],
            "warmup_steps": [814],
        },
        "race": {
            "learning_rate": [2e-5],
            "per_device_train_batch_size": [32],
            "attention_probs_dropout_prob": [0.1],
            "classifier_dropout_prob": [0.1],
            "max_steps": [12000],
            "warmup_steps": [1000],
        },
    }

    # To finetune the pre-trained models, we search the optimization hyperparameters
    # in a search space including different batch sizes (16/32/48), learning
    # rates ((1-10) * e-5), and the number of epochs (2-10)
    from ..result_analysis.azure_utils import JobID
    return get_space_union_and_unique(search_space_common, search_space_unique,
                                      [JobID.get_full_data_name(
                                          dataset_name_list,
                                          subdataset_name)])
